# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

# from wield.iirrational import fitters
from wield.iirrational import fitters_rational
from wield.iirrational import plots
from wield.iirrational.utilities import relpy
import numpy as np
from os import path

try:
    from IIRrational_test_data import IIRrational_data_dev
    from IIRrational_test_data import matlab as matlab_test_data
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False

localpath = path.split(__file__)[0]
skipdeco = pytest.mark.skipif(
    module_import_skip, reason="cannot import IIRrational_test_data"
)


@skipdeco
def test_imcheby_convergence():
    data = IIRrational_data_dev("quad_M0_P2V_10Hz")
    SNR = np.minimum(data.SNR, 100)
    fitter = fitters_rational.ChebychevFilter(
        F_Hz=data.F_Hz,
        data=data.data,
        W=SNR,
        npoles=100,
        nzeros=100,
    )
    print("data", fitter.data)
    print("W", fitter.W)
    print("h_a", fitter.h_a.xfer, fitter.h_a.lnG)
    print("h_b", fitter.h_b.xfer, fitter.h_b.lnG)
    fitter.fit_zeros()
    fitter.fit_poles()
    print("h_a", fitter.h_a.xfer[0], fitter.h_a.lnG)
    print("h_b", fitter.h_b.xfer[0], fitter.h_b.lnG)
    fitter.fit_zeros()
    print("h_a", fitter.h_a.xfer[0], fitter.h_a.lnG)
    print("h_b", fitter.h_b.xfer[0], fitter.h_b.lnG)
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.3)
    fitter.stabilize = True
    # fitter.npoles = len(fitter.poles)
    # fitter.nzeros = len(fitter.zeros)
    ##fitter.stabilize = True
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.7)

    fitter2 = fitters_rational.ChebychevFilter(
        F_Hz=fitter.F_Hz,
        data=fitter.data,
        W=fitter.W,
    )
    # fitter2.a_vec = fitter.a_vec
    # fitter2.b_vec = fitter.b_vec
    # fitter2.gain = fitter.gain
    # fitter2.poles = fitter.poles
    # fitter2.zeros = fitter.zeros
    # fitter2.gain = fitter.gain

    # TODO
    #!Make this a TEST so it doesn't plot all of the time

    axB = plots.plot_fit(
        fitter,
        plot_zp=True,
        # plot_past_data = False,
        plot_past_data=True,
        xscale="log",
    )

    axB.save(relpy(__file__, "plots/imcheby-convergence"))
    return


@skipdeco
def test_imcheby_convergence_2():
    data = IIRrational_data_dev("quad_M0_P2V_10Hz")
    SNR = np.minimum(data.SNR, 100)
    fitter = fitters_rational.ChebychevFilter(
        F_Hz=data.F_Hz,
        data=data.data,
        W=SNR,
        npoles=20,
        nzeros=18,
    )
    print("data", fitter.data)
    print("W", fitter.W)
    print("h_a", fitter.h_a.xfer, fitter.h_a.lnG)
    print("h_b", fitter.h_b.xfer, fitter.h_b.lnG)
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    ##print("HOLD")
    fitter.npoles = 62
    fitter.nzeros = 60
    ####print('h_a', fitter.h_a.xfer[0], fitter.h_a.lnG)
    ####print('h_b', fitter.h_b.xfer[0], fitter.h_b.lnG)
    fitter.fit_zeros()
    ####print('h_a', fitter.h_a.xfer[0], fitter.h_a.lnG)
    ####print('h_b', fitter.h_b.xfer[0], fitter.h_b.lnG)
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.3)
    fitter.stabilize = True
    ####fitter.npoles = len(fitter.poles)
    ####fitter.nzeros = len(fitter.zeros)
    #####fitter.stabilize = True

    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.7)
    # fitter.fit_poles()
    # fitter.fit_zeros()

    # TODO
    #!Make this a TEST so it doesn't plot all of the time

    axB = plots.plot_fit(
        fitter,
        plot_zp=True,
        plot_past_data=False,
        # plot_past_data = True,
        xscale="log",
    )
    axB.save(relpy(__file__, "plots/imcheby-convergence_2"))

    axB = plots.plot_fit(
        fitter,
        plot_zp=True,
        # plot_past_data = False,
        plot_past_data=True,
        xscale="log",
    )

    axB.save(relpy(__file__, "plots/imcheby-convergence_3"))
    return


@skipdeco
def test_imcheby_convergence_alpha():
    fname = matlab_test_data.matfiles["alphas_fdomain.txt"]
    alphas = np.loadtxt(fname)
    # assemble back as complex numbers (each column of a is a TF)
    fr = alphas[:, 0]
    a = np.zeros((alphas.shape[0], (alphas.shape[1] - 1) // 2), dtype=complex)
    for i in range(a.shape[1]):
        a[:, i] = alphas[:, i + 1] + 1j * alphas[:, i + 2]

    idx = np.where(np.logical_and(fr > 30, fr < 300))[0]

    order = 20
    fitter = fitters_rational.ChebychevFilter(
        data=a[idx, 0],
        F_Hz=fr[idx],
        W=10,
        zeros_overlay=(-1,),
        poles_overlay=(-30,),
        npoles=order // 2,
        nzeros=order // 2,
    )
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.nzeros = order
    fitter.npoles = order
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.5)
    fitter.stabilize = True
    # fitter.npoles = len(fitter.poles)
    # fitter.nzeros = len(fitter.zeros)
    ##fitter.stabilize = True
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.5)
    # fitter.stabilize = False
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.1)

    print("Z: ", fitter.ZPKrep.zeros)
    print("P: ", fitter.ZPKrep.poles)

    # from wield.iirrational import fitters_ZPK
    # fitter = fitters_ZPK.ZPKrep2MRF(
    #    ZPKrep = fitter.ZPKrep,
    #    residuals_type      = 'log',
    #    coding_map          = fitters_ZPK.codings_s.coding_maps.SOSnl,
    #    #coding_map          = fitters_ZPK.codings_s.coding_maps.nlFBW,
    #    distance_limit_auto = 0,
    #    max_BW_Hz           = np.max(fitter.F_Hz),
    # )
    # TODO
    #!Make this a TEST so it doesn't plot all of the time

    axB = plots.plot_fit(
        fitter,
        plot_zp=True,
        # plot_past_data = False,
        plot_past_data=True,
        xscale="log",
    )

    axB.save(relpy(__file__, "plots/imcheby-convergence"))
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Calculation of the RMS of a filter with unit white noise passing through a ZPK.
This is done using an integral and residue calculus. The filter must have more
P1 than Z, unless the ZPK is in the Z domain,
where there is a natural cutoff frequency.
"""

import numpy as np

import numpy.polynomial.chebyshev as cheby
import collections


def ZPKreciprocal(ZPK):
    Z, P, K = ZPK
    return P, Z, 1 / K


def ZPKscalarprod(ZPK, val):
    Z, P, K = ZPK
    return Z, P, K * val


def ZPKprod(ZPK1, ZPK2):
    Z1, P1, K1 = ZPK1
    Z2, P2, K2 = ZPK2
    return (
        tuple(Z1) + tuple(Z2),
        tuple(P1) + tuple(P2),
        K1 * K2,
    )


def ZPKdiv(ZPK1, ZPK2):
    Z1, P1, K1 = ZPK1
    Z2, P2, K2 = ZPK2
    return (
        tuple(Z1) + tuple(P2),
        tuple(P1) + tuple(Z2),
        K1 / K2,
    )


def ZPKscalardiv(ZPK, val):
    Z, P, K = ZPK
    return Z, P, K / val


def ZPKdivscalar(val, ZPK):
    Z, P, K = ZPK
    return P, Z, val / K


def ZPKscalarsum(ZPK, val):
    return ZPKsum(ZPK, ((), (), val))


def ZPKsum(ZPK1, ZPK2, scale=None):
    Z1, P1, K1 = ZPK1
    Z2, P2, K2 = ZPK2

    if scale is None:
        # TODO, check for weak high-frequency zeros?
        # TODO, use newer numpy "initial" argument?
        scale = 1 / max(
            np.amax(np.concatenate([np.abs(Z1), [0]])),
            np.amax(np.concatenate([np.abs(P1), [0]])),
            np.amax(np.concatenate([np.abs(Z2), [0]])),
            np.amax(np.concatenate([np.abs(P2), [0]])),
        )

    # apply now before we rescale the poles
    P3 = tuple(P1) + tuple(P2)

    Z1 = -1j * np.asarray(Z1) * scale
    P1 = -1j * np.asarray(P1) * scale
    # Not rescaling K1, even though we should, it would have to be inverted later
    Z2 = -1j * np.asarray(Z2) * scale
    P2 = -1j * np.asarray(P2) * scale
    # Not rescaling K2, even though we should, it would have to be inverted later

    b1 = K1 * (-1j * scale) ** (len(P1) - len(Z1)) * cheby.chebfromroots(Z1)
    a1 = cheby.chebfromroots(P1)

    b2 = K2 * (-1j * scale) ** (len(P2) - len(Z2)) * cheby.chebfromroots(Z2)
    a2 = cheby.chebfromroots(P2)

    b3_a = cheby.chebmul(b1, a2)
    b3_b = cheby.chebmul(b2, a1)
    b3 = cheby.chebadd(b3_a, b3_b)

    # print(b1, a2)
    # print(b2, a1)
    # print(b3)

    Z3 = 1j * cheby.chebroots(b3) / scale
    K3 = b3[-1] / (-1j * scale) ** (len(P3) - len(Z3)) * 2 ** (len(b3) - 2)

    # TODO, CHECK THIS, needs test cases
    return tuple(Z3), tuple(P3), K3



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Calculation of the RMS of a filter with unit white noise passing through a ZPK.
This is done using an integral and residue calculus. The filter must have more
poles than zeros, unless the ZPK is in the Z domain,
where there is a natural cutoff frequency.
"""

import numpy as np

from scipy.special import factorial
from scipy import signal as scisig
import numpy.polynomial.chebyshev as cheby


class ChebyResiduesHelper(object):
    _ZPK = None

    def __init__(
        self,
        ZPK,
        scale=None,
        rotate=True,
    ):
        # store for the setup, will be cleared after setup has used it
        self._ZPK = ZPK
        self.scale = scale

        if rotate:
            self.rot_p = +1j
            self.rot_n = -1j
        else:
            self.rot_p = 1
            self.rot_n = 1

    def _setup(self):
        if self._ZPK is None:
            return

        zeros, poles, gain = self._ZPK
        # clear the _ZPK holder to indicate that setup has been fully run
        self._ZPK = None
        if self.scale is None:
            self.scale = 1 / max(
                np.amax(np.concatenate([np.abs(zeros), [0]])),
                np.amax(np.concatenate([np.abs(poles), [0]])),
            )
        # print('SCAL:', self.scale)

        self.zeros = self.rot_n * np.asarray(zeros) * self.scale
        self.poles = self.rot_n * np.asarray(poles) * self.scale
        # technically should rescale K here as well, but not, so that it doesn't
        # have to just be inverted later

        b = gain * cheby.chebfromroots(self.zeros)
        self.a = cheby.chebfromroots(self.poles)
        # print("b orig:", cheby.chebfromroots(self.zeros))
        # print("a orig:", self.a)
        self.orderdiff = len(b) - len(self.a)

        k, self.b = cheby.chebdiv(b, self.a)
        self.k = k * 2 ** (self.orderdiff - 1)
        return

    _k_zeros = None

    @property
    def k_zeros(self):
        if self._k_zeros is None:
            self._setup()
            self._k_zeros = self.rot_p * cheby.chebroots(self.k) / self.scale
        return self._k_zeros

    @property
    def k_gain(self):
        self._setup()
        # no need to rescale k here since we didn't rescale it originally
        return self.k[-1]

    def residue_of(self, polemask):
        """
        Takes an untransformed pole, transforms it
        """
        self._setup()
        # rescale pole
        residue_list = []
        bn = self.b.copy()

        an = np.atleast_1d(cheby.chebfromroots(self.poles[polemask]))

        pavg = np.average(self.poles[~polemask])
        # bn(s) / an(s) is (s-po[n])**Nn * b(s) / a(s) where Nn is
        # multiplicity of pole at po[n]
        mult = np.count_nonzero(~polemask)
        for m in range(mult, 0, -1):
            if mult > m:
                # compute next derivative of bn(s) / an(s)
                term1 = cheby.chebmul(cheby.chebder(bn, 1), an)
                term2 = cheby.chebmul(bn, cheby.chebder(an, 1))
                bn = cheby.chebsub(term1, term2)
                an = cheby.chebmul(an, an)
            val = (
                self.rot_p ** (m + self.orderdiff)
                * (
                    (cheby.chebval(pavg, bn) / cheby.chebval(pavg, an))
                    / factorial(mult - m)
                )
                / (self.scale) ** (m + self.orderdiff)
            )

            residue_list.append(val)
        residue_list.reverse()
        return residue_list


def ZPK2residues(
    ZPK,
    F_nyquist_Hz=None,
    Npoles=None,
    _helper_order=None,  # for debugging the derivatives
):
    """
    Npoles allows one to compute residues for fewer poles than given,
    for instance to compute them only for the inner unit circle for ZPKs
    representing power-spectra.
    """
    # currently fills this
    residue_list = []

    zeros, poles, gain = ZPK
    zeros = np.asarray(zeros)
    poles = np.asarray(poles)

    if Npoles is None:
        Npoles = len(poles)

    helper = ChebyResiduesHelper(ZPK)

    # list of enumerated poles that will be depleted as the integral is computed
    # over all residues. Depletion is used since some poles may be stacked
    ppairs = dict(enumerate(poles[:Npoles]))

    res_thresh = 1e-8
    while ppairs:
        idx, p = ppairs.popitem()

        p_eval = p - poles
        pmask = abs(p_eval) > res_thresh
        D_0d = np.prod(p_eval[pmask])
        Nd = np.count_nonzero(~pmask) - 1

        for idx_bad in np.argwhere(~pmask[:Npoles]):
            # remove the multiplicities from the pool to compute
            if idx_bad == idx:
                continue
            # must convert bad to standard python type for the pop to index
            idx_bad = int(idx_bad)
            ppairs.pop(idx_bad)

        if _helper_order is not None and Nd >= _helper_order or Nd >= 4:
            res = helper.residue_of(pmask)
            residue_list.append(
                (p,) + tuple(res),
            )
            # skip the calculation below if already computed
            continue

        N_0d = eval_Nth_derivative(p, zeros, Nd=0)
        R = N_0d / D_0d
        eval_0d = gain * R

        if Nd >= 1:
            # TODO, make the derivatives work up to some order
            N_1d = eval_Nth_derivative(p, zeros, val=N_0d, Nd=1)
            D_1d = eval_Nth_derivative(p, poles, val=D_0d, rootmask=pmask, Nd=1)
            R_1d = (N_1d - D_1d * N_0d / D_0d) / D_0d
            eval_1d = gain * R_1d

        if Nd >= 2:
            N_2d = eval_Nth_derivative(p, zeros, val=N_0d, Nd=2)
            D_2d = eval_Nth_derivative(p, poles, val=D_0d, rootmask=pmask, Nd=2)
            # TODO, make the derivatives work up to some order
            R_2d = (N_2d - D_2d * N_0d / D_0d - 2 * R_1d * D_1d) / D_0d
            eval_2d = gain * R_2d / 2

        if Nd >= 3:
            # TODO, make the derivatives work up to some order
            N_3d = eval_Nth_derivative(p, zeros, val=N_0d, Nd=3)
            D_3d = eval_Nth_derivative(p, poles, val=D_0d, rootmask=pmask, Nd=3)
            # TODO, make the derivatives work up to some order
            R_3d = (
                N_3d
                + (
                    -D_1d * N_2d
                    - D_2d * N_1d
                    - D_3d * N_1d
                    + 2 * R_1d * D_1d ** 2
                    + 2 * N_0d * D_2d * D_1d / D_0d
                )
                / D_0d
                - 2 * (R_2d * D_1d + R_1d * D_2d)
            ) / D_0d
            eval_3d = gain * R_3d / 6

        if Nd == 0:
            residue_list.append(
                (p, eval_0d),
            )
        elif Nd == 1:
            residue_list.append(
                (p, eval_1d, eval_0d),
            )
        elif Nd == 2:
            residue_list.append(
                (p, eval_2d, eval_1d, eval_0d),
            )
        elif Nd == 3:
            residue_list.append(
                (p, eval_3d, eval_2d, eval_1d, eval_0d),
            )
        else:
            raise RuntimeError(
                "Residues with poles of mult>3 not implemented"
                " with direct derivatives"
            )

    if len(zeros) < len(poles):
        return residue_list, ((), (), 0)
    elif len(zeros) == len(poles):
        return residue_list, ((), (), gain)
    else:
        return residue_list, (
            tuple(helper.k_zeros),
            (),
            np.asarray(helper.k_gain).item(),
        )


def eval_Nth_derivative(z, roots, rootmask=None, val=None, Nd=0):
    if rootmask is None:
        rootmask = np.ones(roots.shape, dtype=bool)

    if Nd == 0:
        if val is None:
            return np.prod(z - roots[rootmask])
        else:
            return val

    if val is None and Nd < 3:
        r_eval = z - roots[rootmask]
        val = np.prod(r_eval)

    # TODO, currently this algorithm is very inefficient
    # for large derivatives, it should be able to factor in
    # multiplicities by only iterating indices ABOVE the first zero mask.
    d_sum = 0
    for idx in np.argwhere(rootmask):
        mask_down = np.copy(rootmask)
        mask_down[idx] = 0
        if val is not None:
            r_eval = z - roots[idx]
            # check numerical stability, don't divide out small numbers
            if abs(r_eval) > 1e-8:
                val_down = val / r_eval
            else:
                val_down = None
        else:
            val_down = None
        d_sum += eval_Nth_derivative(z, roots, mask_down, val=val_down, Nd=Nd - 1)
    return np.asarray(d_sum).item()


def ZPK2residues_scipy(
    ZPK,
    F_nyquist_Hz=None,
):
    """
    Useful for testing
    """
    residue_list = []
    b, a = scisig.zpk2tf(*ZPK)
    if F_nyquist_Hz is None:
        r, p, k = scisig.residue(b, a)
    else:
        r, p, k = scisig.residuez(b, a)

    # some indexing tricks to change the output to the form above
    same = p[1:] == p[:-1]
    same_adj = np.concatenate([[False], same, [False]])
    same_shift = ~(same_adj[1:] | same_adj[:-1])
    same_starts = ~same_adj[:-1] & same_adj[1:]
    same_ends = same_adj[:-1] & ~same_adj[1:]
    # the addition of same_adj puts all of the non-multiple back in
    for idx_start, idx_end in zip(
        np.argwhere(same_starts | same_shift), np.argwhere(same_ends | same_shift)
    ):
        idx_start = int(idx_start)
        idx_end = int(idx_end)
        pole = p[idx_start]
        res = r[idx_start : idx_end + 1]
        residue_list.append((pole,) + tuple(res))

    if len(k) < 2:
        return residue_list, ((), (), k[0])
    else:
        Z, P, K = scisig.tf2zpk(k, [1])
        return residue_list, (tuple(Z), tuple(P), K)


def residues2ZPK(rlist, ZPKadd=None, scale=None, rotate=True):
    """ """
    poles = []
    for rtup in rlist:
        pole = rtup[0]
        poles.extend([pole] * (len(rtup) - 1))
    poles = np.asarray(poles)

    if scale is None:
        scale = 1 / np.amax(np.concatenate([abs(poles), [0]]))

    if rotate:
        rot_p = +1j
        rot_n = -1j
    else:
        rot_p = 1
        rot_n = 1

    # scale = .5
    # print(scale)
    poles_scaled = rot_n * poles * scale

    if ZPKadd is None:
        b = [0]
        orderdiff = 1 - len(poles)
    else:
        Z, P, K = ZPKadd
        Z = rot_n * np.asarray(Z) * scale
        b = cheby.chebmul(cheby.chebfromroots(poles_scaled), cheby.chebfromroots(Z) * K)

        orderdiff = len(b) - len(poles) - 2
        if len(P):
            raise NotImplementedError("Currently Cannot Sum ZPK with poles")
    idx_plist = 0
    # print("orderdiff", orderdiff)
    # print("b: start ", b)
    for rtup in rlist:
        pole = poles_scaled[idx_plist]
        mult = len(rtup) - 1
        poles_temp = np.concatenate(
            [
                poles_scaled[:idx_plist],
                poles_scaled[idx_plist + mult :],
            ]
        )
        a_outer = cheby.chebfromroots(poles_temp)

        for m, res in enumerate(rtup[1:]):
            a_inner = cheby.chebfromroots([pole] * (mult - m - 1))
            a_temp = cheby.chebmul(a_outer, a_inner)
            # TODO, might be able to avoid the scaling factor here if scaling
            # of the original b is done differently at the beginning
            res_scaled = res * rot_n ** (m - orderdiff) * scale ** (2 + orderdiff + m)
            b = cheby.chebadd(b, res_scaled * a_temp)
            # print(res_scaled * a_temp)
        idx_plist += mult

    # print(b)
    gain = b[-1] * 2 ** (len(b) - 2)  # / scale**len(b)
    zeros = rot_p * cheby.chebroots(b) / scale

    return (tuple(zeros), tuple(poles), gain)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np


def squeezerec(key, obj):
    if isinstance(obj, dict):
        for k, v in list(obj.items()):
            obj[k] = squeezerec(k, v)
        return obj
    elif isinstance(obj, np.ndarray):
        if obj.dtype.names is None:
            if not obj.shape:
                return squeezerec(None, obj.item())
            if obj.shape == (0,) or obj.shape == (0, 0):
                return None
            else:
                if obj.dtype == object:
                    lst = []
                    for idx, v in enumerate(obj):
                        lst.append(squeezerec(idx, v))
                    return np.array(lst)
                else:
                    return obj.squeeze()
        else:
            d = dict()
            for name in obj.dtype.names:
                d[name] = squeezerec(name, obj[name])
            return d
    else:
        return obj


def load_matlab(fname):
    import scipy.io

    d = dict()
    d = scipy.io.loadmat(
        fname,
        mdict=d,
        squeeze_me=True,
        chars_as_strings=True,
        # mat_dtype = True,
    )
    d = squeezerec(None, d)
    d.pop("__globals__", None)
    d.pop("__header__", None)
    d.pop("__version__", None)
    return d


def desqueezerec(key, obj):
    if obj is None:
        return np.array([])
    elif isinstance(obj, dict):
        # convert to recarray?
        for k, v in list(obj.items()):
            obj[k] = desqueezerec(k, v)
        return obj
    else:
        return obj


def write_matlab(fname, obj):
    import scipy.io

    d = scipy.io.savemat(
        fname,
        mdict=desqueezerec(None, obj),
    )
    return d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

from wield import declarative
import collections

NOARG = declarative.unique_generator


def subkey_search(fdict, subkey, default=NOARG):
    if subkey is None:
        return fdict
    subdict = fdict
    skeys = subkey.split(".")

    while skeys:
        if not isinstance(subdict, collections.Mapping):
            raise TypeError("Intermediate type not a dictionary")

        try:
            subdict = subdict[".".join(skeys)]
        except KeyError:
            pass
        else:
            break

        for idx in range(1, len(skeys)):
            semikey = ".".join(skeys[:-idx])
            try:
                subdict = subdict[semikey]
            except KeyError:
                pass
            else:
                skeys = skeys[-idx:]
                break
        else:
            if default is NOARG:
                # this is the for-else syntax, only triggered if the for loop never
                # finds a match
                raise KeyError("Could not recursively find {}".format(subkey))
            else:
                return default
    return subdict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
import scipy
import scipy.linalg

from wield.bunch import Bunch

from .data_filtfit_base import DataFiltFitBase
from .roots_bin import roots_bin_palindromicz, roots_re_pair
from .svd import SVD_SV


def abs_sq(x):
    return x.real ** 2 + x.imag ** 2


class RationalDiscFilterMag(DataFiltFitBase):
    phase_missing = False

    def __build__(
        self,
        _args=None,
        npoles=None,
        nzeros=None,
        ZPK=((), (), 1),
        parent=None,
        **kwargs
    ):
        if _args:
            raise RuntimeError("Only keyword arguments allowed")
        self.npoles = npoles
        self.nzeros = nzeros

        super(RationalDiscFilterMag, self).__build__(parent=parent, **kwargs)

        # first to prevent lookups that shouldn't happen
        self.hsq_b_ref = 1
        self.dependencies_for("zeros", "bsq_vec")
        self.hsq_a_ref = 1
        self.dependencies_for("poles", "asq_vec")
        self.zeros = ZPK[0]
        self.poles = ZPK[1]

        self.gain_sq = 1

        @self.deco_generator()
        def data_magsq(self):
            return abs_sq(self.data)

        @self.deco_generator(clear=False)
        def asq_vec(self):
            """
            Setup so that the gain at the reference freq is ALWAYS 1
            """
            vec = self.poly.fromroots(self.poles).real
            asq_vec_full = np.convolve(vec, vec[::-1] / vec[0], mode="full")
            asq_vec = asq_vec_full[len(vec) - 1 :]
            self.dependencies("poles")
            return asq_vec

        @self.deco_setter(clear=False)
        def asq_vec(self, val):
            gain_sq = self.gain_sq
            self.gain_sq = gain_sq / abs(val[-1])
            val = val / abs(val[-1])
            self.dependencies("poles")
            return val

        @self.deco_generator(clear=False)
        def bsq_vec(self):
            """
            Setup so the gain is always specified at the ref
            """
            # TODO fix gain deps
            vec = self.get_raw("gain") * self.poly.fromroots(self.zeros).real
            bsq_vec_full = np.convolve(vec, vec[::-1] / vec[0], mode="full")
            bsq_vec = bsq_vec_full[len(vec) - 1 :]
            self.dependencies("zeros")
            return bsq_vec

        @self.deco_setter(clear=False)
        def bsq_vec(self, val):
            val = np.asarray(val)
            self.gain_sq = abs(val[-1])
            val = val / abs(val[-1])
            self.dependencies("zeros")
            return val

        @self.deco_generator(clear=False)
        def zeros_split(self):
            vec = np.concatenate([self.bsq_vec[1::][::-1], self.bsq_vec])
            return roots_bin_palindromicz(
                self.poly.roots(vec), F_nyquist_Hz=self.F_nyquist_Hz
            )

        @self.deco_generator(clear=False)
        def zeros(self):
            self.dependencies("bsq_vec")
            return roots_re_pair(*self.zeros_split)

        @self.deco_setter(clear=False)
        def zeros(self, val):
            # get previous
            val = np.asarray(val)
            self.dependencies("bsq_vec")
            self.hsq_b_ref = self.poly.valfromroots(self.Xp_ref, val)
            return val

        @self.deco_generator
        def h_b_ref(self):
            return self.hsq_b_ref ** 0.5

        @self.deco_generator
        def h_a_ref(self):
            return self.hsq_a_ref ** 0.5

        @self.deco_generator(clear=False)
        def poles_split(self):
            vec = np.concatenate([self.asq_vec[1::][::-1], self.asq_vec])
            r_r, r_c = roots_bin_palindromicz(
                self.poly.roots(vec), F_nyquist_Hz=self.F_nyquist_Hz
            )
            return r_r, r_c

        @self.deco_generator(clear=False)
        def poles(self):
            self.dependencies("asq_vec")
            return roots_re_pair(*self.poles_split)

        @self.deco_setter(clear=False)
        def poles(self, val):
            val = np.asarray(val)
            self.dependencies("asq_vec")
            h_ref_now = abs_sq(self.poly.valfromroots(self.Xp_ref, val))
            h_ref_prev = self.hsq_a_ref
            # self.gain = self.get_raw('gain') / abs(h_ref_prev / h_ref_now)**.5
            self.hsq_a_ref = h_ref_now
            return val

        @self.deco_generator
        def poles_full(self):
            return np.asarray(self.poles)

        @self.deco_generator
        def zeros_full(self):
            return np.asarray(self.zeros)

        @self.deco_setter(clear=True)
        def gain(self, val):
            self.dependencies("bsq_vec")
            return val

        @self.deco_generator(clear=True)
        def gain(self):
            self.dependencies("bsq_vec")
            # TODO, could make more stable
            gain2 = (
                np.product(np.abs(self.poles)) / np.product(np.abs(self.zeros))
            ) ** 0.5
            return self.gain_sq ** 0.5 * gain2

        self.gain = ZPK[2]

        @self.deco_generator(clear=False)
        def V_b(self):
            # return self.poly.vander(self.Xc_grid, self.nzeros).real
            # TODO there is a more efficient way to construct this
            v = self.poly.vander(self.Xp_grid, self.nzeros).real
            v[:, 1:] *= 2
            return v

        @self.deco_generator(clear=False)
        def V_a(self):
            # return self.poly.vander(self.Xc_grid, self.npoles).real
            # TODO there is a more efficient way to construct this
            v = self.poly.vander(self.Xp_grid, self.npoles).real
            v[:, 1:] *= 2
            return v

        @self.deco_generator
        def V_ref_b(self):
            v = self.poly.vander(self.Xp_ref, self.nzeros).real
            v[:, 1:] *= 2
            return v

        @self.deco_generator
        def V_ref_a(self):
            v = self.poly.vander(self.Xp_ref, self.npoles).real
            v[:, 1:] *= 2
            return v

        @self.deco_generator(clear=False)
        def h_a(self):
            # TODO account for phase adjust
            self.dependencies("poles", "asq_vec")
            val = self.poly.valfromroots(self.Xp_grid, self.poles)
            return val

        @self.deco_generator(clear=False)
        def h_b(self):
            # TODO account for phase adjust
            self.dependencies("zeros", "bsq_vec", "gain")
            val = self.poly.valfromroots(self.Xp_grid, self.zeros)
            return self.gain * val

        @self.deco_generator(clear=False)
        def hsq_a(self):
            self.dependencies("poles", "asq_vec")
            return np.dot(self.V_a[:, : len(self.asq_vec)], self.asq_vec)

        @self.deco_generator(clear=False)
        def hsq_b(self):
            # TODO account for zeros_phase_adjust
            self.dependencies("zeros", "bsq_vec", "gain")
            return np.dot(self.V_b[:, : len(self.bsq_vec)], self.bsq_vec)

        @self.deco_generator(clear=False)
        def xfer_fit(self):
            return self.xfer_fit_magsq ** 0.5
            return self.h_b / self.h_a

        @self.deco_generator(clear=False)
        def xfer_fit_magsq(self):
            return self.gain_sq * self.hsq_b / self.hsq_a

        @self.deco_generator(clear=False)
        def residuals(self):
            debias_reweight = 1 / (0.001 + self.W ** 2)
            retB = Bunch()
            R = self.xfer_fit_magsq / self.data_magsq
            retB.resP = self.W * (R - 1) / 2
            retB.resZ = self.W * (1 / R - 1) / 2
            retB.resD = self.W * (R - 1 / R) / 2
            retB.resD_average = np.sum(abs(retB.resD) ** 2) / (4 * len(self.data))
            retB.average = np.sum(
                (abs(retB.resP) ** 2 + abs(retB.resZ * debias_reweight) ** 2)
                / (1 + debias_reweight)
            ) / (2 * len(self.data))
            return retB

        @self.deco_generator(clear=False)
        def residuals_average(self):
            return self.residuals.resD_average

        @self.deco_generator(clear=False)
        def A_z(self):
            return self.V_b * (self.W / (self.data_magsq * self.hsq_a)).reshape(-1, 1)

        @self.deco_generator(clear=False)
        def A_zp(self):
            return self.V_a * (
                self.W * self.hsq_b / (self.data_magsq * self.hsq_a ** 2)
            ).reshape(-1, 1)

        @self.deco_generator(clear=False)
        def A_p(self):
            return self.V_a * (self.W * (self.data_magsq / self.hsq_b)).reshape(-1, 1)

        @self.deco_generator(clear=False)
        def A_pz(self):
            return self.V_b * (
                self.W * (self.data_magsq * self.hsq_a / self.hsq_b ** 2)
            ).reshape(-1, 1)

        return  # ~__init__

    def matched_pairs_clear(
        self,
        Q_rank_cutoff=0.5,
    ):
        """
        Match unique closest pairs, if they are within a bandwidth of 0Hz, then they are ignored
        """
        poles_r, poles_c = self.poles_split
        zeros_r, zeros_c = self.zeros_split

        def nearest_idx(lst_1, lst_2):
            nearest_lst = []
            for r1 in lst_1:
                if r1 is None:
                    nearest_lst.append(None)
                    continue
                dist_nearest = float("inf")
                idx_nearest = None
                for idx_2, r2 in enumerate(lst_2):
                    if r2 is None:
                        continue
                    dist = abs(r1 - r2)
                    if dist < dist_nearest:
                        idx_nearest = idx_2
                        dist_nearest = dist
                nearest_lst.append(idx_nearest)
            return nearest_lst

        z_nearest = nearest_idx(zeros_c, poles_c)
        p_nearest = nearest_idx(poles_c, zeros_c)

        z_duals = []
        p_duals = []
        for idx_z, idx_p in enumerate(z_nearest):
            if idx_p is None:
                continue
            if idx_z != p_nearest[idx_p]:
                # not a unique pairing
                continue
            z = zeros_c[idx_z]
            p = poles_c[idx_p]
            Q_rank = abs(p - z) * (1 / (1 - abs(p)) ** 2 + 1 / (1 - abs(z)) ** 2) ** 0.5
            if Q_rank < Q_rank_cutoff:
                z_duals.append(idx_z)
                p_duals.append(idx_p)
        p_duals = set(p_duals)
        z_duals = set(z_duals)

        poles_new = []
        for idx_p, pole in enumerate(poles_c):
            if idx_p in p_duals:
                continue
            poles_new.append(pole)
            poles_new.append(pole.conjugate())
        zeros_new = []
        for idx_z, zero in enumerate(zeros_c):
            if idx_z in z_duals:
                continue
            zeros_new.append(zero)
            zeros_new.append(zero.conjugate())
        poles_new.extend(poles_r)
        zeros_new.extend(zeros_r)
        self.poles = poles_new
        self.zeros = zeros_new
        return

    def fit_poles_mod_zeros(self, max_size=None, **kwargs):
        # remove the effect of linear (delay) phasing
        A_z = self.A_z
        # A_z = np.hstack([(1j*F_Hz).reshape(-1, 1), A_z])
        q, r = np.linalg.qr(A_z)

        A_p = self.A_p
        # the str is because py2 doesn't like some symbol in there
        A_p = A_p - np.einsum(
            str("ij,jk->ik"), q, np.einsum(str("ij,ik->jk"), q.conjugate(), A_p)
        )

        S, V = SVD_SV(
            A_p.real,
            n_smallest=3,
            overwrite_a=True,
        )
        # print("POLES SVD: ", S[-4:])
        self.asq_vec = V.T[:, -1]
        return

    def fit_zeros_mod_poles(self, max_size=None, **kwargs):
        A_p = self.A_p
        # remove the effect of linear (delay) phasing
        # A_p = np.hstack([(1j*F_Hz).reshape(-1, 1), A_p])

        q, r = np.linalg.qr(A_p)

        A_z = self.A_z
        A_z = A_z - np.einsum(
            str("ij,jk->ik"), q, np.einsum(str("ij,ik->jk"), q.conjugate(), A_z)
        )
        S, V = SVD_SV(
            A_z.real,
            n_smallest=3,
            overwrite_a=True,
        )
        # print("ZEROS SVD: ", S[-4:])
        self.bsq_vec = V.T[:, -1]
        return

    def match_pair_iter(
        self,
        Q_rank_cutoff,
        num_sequence=2,
        zeros_first=True,
    ):
        npoles = self.npoles
        nzeros = self.nzeros

        self.matched_pairs_clear(Q_rank_cutoff=Q_rank_cutoff)

        self.npoles = len(self.poles)
        self.nzeros = len(self.zeros)

        if not zeros_first:
            self.fit_poles()
            num_sequence -= 1
        for idx in range(num_sequence):
            self.fit_zeros()
            self.fit_poles()

        if not zeros_first:
            self.fit_zeros()

        self.npoles = npoles
        self.nzeros = nzeros

    def fit_poles(self, **kwargs):
        # print(self.asq_vec, self.bsq_vec)
        A_p = self.A_p
        # solve the problem with purely real taps
        a_fit, res, rank, s = scipy.linalg.lstsq(
            A_p.real,
            self.W.real,
        )
        self.asq_vec = a_fit
        return

    def fit_zeros(self, **kwargs):
        A_z = self.A_z
        b_fit, res, rank, s = scipy.linalg.lstsq(
            A_z.real,
            self.W.real,
        )
        self.bsq_vec = b_fit
        return

    def fit_polesX(self, **kwargs):
        # print(self.asq_vec, self.bsq_vec)
        A_p = self.A_p
        # solve the problem with purely real taps
        a_fit, res, rank, s = scipy.linalg.lstsq(
            self.data_magsq.reshape(-1, 1) * A_p.real,
            self.data_magsq * self.W.real,
        )
        self.asq_vec = a_fit
        return

    def fit_zerosX(self, **kwargs):
        A_z = self.A_z
        b_fit, res, rank, s = scipy.linalg.lstsq(
            1 / self.data_magsq.reshape(-1, 1) * A_z.real,
            1 / self.data_magsq * self.W.real,
        )
        self.bsq_vec = b_fit
        return

    def remove_doublets(self):
        p_r, p_c = self.poles_split
        z_r, z_c = self.zeros_split

        thresh_var = 1e-2
        thresh_covar = 0.1

        # mapping from zeros to poles, inside bunch with covar data
        used_zs = dict()

        for idx_p, p in enumerate(p_c):
            min_idx_z = None
            min_dist = 1
            for idx_z, z in enumerate(z_c):
                dist = abs_sq(p - z)
                if dist < min_dist:
                    min_idx_z = idx_z
                    min_dist = dist

            if min_idx_z is None:
                continue
            rp, rz, rcov_n = self.generate_ZPK_covar(
                cpoles_seq=[idx_p], czeros_seq=[min_idx_z]
            )
            if rp > thresh_var and rz > thresh_var and rcov_n > thresh_covar:
                if min_idx_z in used_zs:
                    if rcov_n < used_zs[min_idx_z].rcov_n:
                        # should maybe find second closest
                        continue

                # print(rp, rz, rcov_n, min_dist**.5)
                used_zs[min_idx_z] = Bunch(
                    idx_p=idx_p,
                    reff=(rp ** -2 + rz ** -2) ** (-0.5),
                    rp=rp,
                    rz=rz,
                    rcov_n=rcov_n,
                )
        drop_ps = []
        drop_zs = []
        for idx_z, pB in used_zs.items():
            drop_zs.append(idx_z)
            drop_ps.append(pB.idx_p)

        drop_ps.sort()
        drop_zs.sort()

        def drop_join(roots, drop):
            c_r_join = []
            idx_r_prev = -1
            for idx_r in drop:
                c_r_join.append(roots[idx_r_prev + 1 : idx_r])
                idx_r_prev = idx_r
            c_r_join.append(roots[idx_r_prev + 1 :])
            from itertools import chain

            return tuple(chain(*c_r_join))

        p_c_new = drop_join(p_c, drop_ps)
        z_c_new = drop_join(z_c, drop_zs)
        self.poles = tuple(p_r) + p_c_new + tuple(r.conjugate() for r in p_c_new)
        self.zeros = tuple(z_r) + z_c_new + tuple(r.conjugate() for r in z_c_new)
        self.npoles = len(self.poles)
        self.nzeros = len(self.zeros)
        return

    def fit_pzpz(
        self,
        max_size=None,
        collect_all=False,
        zeros_first=False,
        n_svd=1,
        n_iter=10,
    ):
        collection = []
        if not zeros_first:
            if n_svd >= 1:
                fitA = self.fit_poles_mod_zeros
            else:
                fitA = self.fit_poles
            if n_svd >= 2:
                fitB = self.fit_zeros_mod_poles
            else:
                fitB = self.fit_zeros
            fitC = self.fit_poles
            fitD = self.fit_zeros
        else:
            if n_svd >= 1:
                fitA = self.fit_zeros_mod_poles
            else:
                fitA = self.fit_zeros
            if n_svd >= 2:
                fitB = self.fit_poles_mod_zeros
            else:
                fitB = self.fit_poles
            fitC = self.fit_zeros
            fitD = self.fit_poles

        fitA(
            max_size=max_size,
        )
        fitB(
            max_size=max_size,
        )

        for i in range(n_iter - 1):
            fitC()
            fitD()
            if collect_all:
                collection.append(self.copy())

        if n_iter > 0:
            fitC()
            fitD()
        if collect_all:
            collection.append(self.copy())
        return collection

    def fit_pz(self, n_iter=0, n_svd=1, **kwargs):
        return self.fit_pzpz(n_iter=n_iter, n_svd=n_svd, **kwargs)

    def xfer_eval(self, F_Hz):
        # TODO account for phase difference
        X = np.exp(1j * np.pi * F_Hz / self.F_nyquist_Hz)
        h_val = self.poly.valfromroots(X, self.zeros) / self.poly.valfromroots(
            X, self.poles
        )
        return self.gain * h_val



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import matplotlib as mpl
from matplotlib import gridspec
import matplotlib.pyplot as plt
from wield import bunch

from .autoniceplot import (
    asavefig,
    patch_axes,
)


def hide_xlabels(ax):
    for tl in ax.get_xticklabels():
        tl.set_visible(False)


def attach_finalizer(ax):
    ax.finalizers = []
    ax.finalized = False

    def finalize():
        ax.finalized = True
        for f in ax.finalizers:
            f()

    ax.finalize = finalize
    return


def generate_ax(ax_group=None):
    ax = bunch.DeepBunch()
    attach_finalizer(ax)
    if ax_group is not None:
        ax_group.finalizers.append(ax.finalize)
    return ax


def generate_stacked_plot_ax(
    name_use_list,
    gs_base=gridspec.GridSpec(1, 1)[0],
    heights_phys_in_default=1.5,
    heights_phys_in={},
    height_ratios={},
    width_ratios=[1],
    xscales="linear",
    xlim=None,
    wspacing=0.04,
    width_phys_in=None,
    fig=None,
    ax_group=None,
    hspace=0.2,
    **kwargs
):
    view_names = []
    height_ratio_list = []
    height_phys_in = 0
    autocalls = dict()
    for ntup in name_use_list:
        if isinstance(ntup, (tuple, list)):
            name, b_use = ntup
        else:
            name = ntup
            b_use = True
        if callable(name):
            autocalls[name] = b_use
            name = name.__name__
        else:
            if not b_use:
                continue
        height_ratio = height_ratios.get(name, 1)
        height_ratio_list.append(height_ratio)
        height_phys_in += height_ratio * heights_phys_in.get(
            name, heights_phys_in_default
        )
        view_names.append(name)

    gs_DC = gridspec.GridSpecFromSubplotSpec(
        len(view_names),
        len(width_ratios),
        subplot_spec=gs_base,
        height_ratios=height_ratio_list,
        width_ratios=width_ratios,
        wspace=wspacing,
        hspace=hspace,
    )

    if not isinstance(xscales, (list, tuple)):
        xscales = [xscales] * len(width_ratios)

    if width_phys_in is None:
        width_phys_in = mpl.rcParams["figure.figsize"][0]

    if fig is None:
        fig = plt.figure()
        fig.set_size_inches(width_phys_in, height_phys_in)
        # fig.set_dpi(160)

    def hide_finalizer(axB, ax_local):
        def finalize():
            hide_xlabels(ax_local)

        return finalize

    def scale_finalizer(ax_local, scale):
        def finalize():
            ax_local.set_xscale(scale)

        return finalize

    axB = generate_ax(ax_group)
    axB.fig = fig

    def save(rootname, **kwargs):
        axB.finalize()
        axB << asavefig(rootname, **kwargs)

    axB.save = save

    for col_idx in range(len(width_ratios)):
        ax_top = None
        ax_list = []
        for idx, name in enumerate(view_names):
            ax_local = fig.add_subplot(gs_DC[idx, col_idx], sharex=ax_top)
            # patch_axes(ax_local)
            ax_local.grid(b=True)
            ax_local.grid(b=True, which="minor", color=(0.9, 0.9, 0.9), lw=0.5)
            axB["ax{0}_{1}".format(idx, col_idx)] = ax_local
            if col_idx == 0:
                axB["ax{0}".format(idx)] = ax_local
            axB[name + "_{0}".format(col_idx)] = ax_local
            if col_idx == 0:
                axB[name] = ax_local
            if ax_top is None:
                ax_top = ax_local
            ax_list.append(ax_local)
            axB.finalizers.append(scale_finalizer(ax_local, xscales[col_idx]))
            if idx < len(view_names) - 1:
                axB.finalizers.append(hide_finalizer(axB, ax_local))
        axB["ax_bottom_{0}".format(col_idx)] = ax_local
        axB["ax_top_{0}".format(col_idx)] = ax_top
        axB["ax_list_{0}".format(col_idx)] = ax_list
        if col_idx == 0:
            axB["ax_bottom"] = ax_local
            axB["ax_top"] = ax_top
            axB["ax_list"] = ax_list
    for call, cparams in autocalls.items():
        call(axB[call.__name__], **cparams)
    return axB



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Contains setup functions for test data, these are returned in a QuickBunch dictionary, with some annotation about the number of data sets
"""

from . import simple_set_1
from . import random_sets

from ..data2testcase import (
    data2testcase,
    testcase2data,
)

from .utilities import (
    assert_almost_equal,
)

datasets = {}
datasets.update(simple_set_1.datasets)
datasets.update(random_sets.datasets)


def IIRrational_data(name, set_num=0, instance_num=0, **kwargs):
    return datasets[name].generator(sN=set_num, iN=instance_num, **kwargs)


__all__ = [
    'assert_almost_equal',
    'data2testcase',
    'testcase2data',
    'IIRrational_data',
    'datasets',
]



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

from wield.iirrational import v1
from wield.iirrational.testing.utilities import (
    sign_validate_hint,
    sign_validate_and_plot_hint,
)
from wield.iirrational import fitters_rational, fitters_ZPK
from wield.iirrational.testing.plots import plot_on_assert

try:
    from IIRrational_test_data import IIRrational_data_dev
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_L2P_MULT(request, browser, plotsections, plot_verbosity):
    data_name = "L2P_MULT"
    out = v1.data2filter(
        IIRrational_data_dev(
            data_name,
            instance_num=1,
            set_num=2,
        ),
        hints=[sign_validate_and_plot_hint(__file__, request)],
        # order_initial = 80,
    )

    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        pass
    return


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_L2P_MULT_S(request, browser, plotsections, plot_verbosity):
    data_name = "L2P_MULT_WLF"
    out = v1.data2filter(
        IIRrational_data_dev(
            data_name,
            instance_num=1,
            set_num=2,
        ),
        F_nyquist_Hz=None,
        # order_initial = 80,
        hints=[sign_validate_and_plot_hint(__file__, request)],
        delay_s=0,
    )
    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        pass
    return


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_L2P_MULT_S_WOLAY(request, browser, plotsections, plot_verbosity):
    data_name = "L2P_MULT_WLF"
    out = v1.data2filter(
        IIRrational_data_dev(
            data_name,
            instance_num=1,
            set_num=2,
        ),
        F_nyquist_Hz=None,
        order_initial=80,
        delay_s=0,
        ZPK_overlay=((0,), (), 1),
        hints=[sign_validate_and_plot_hint(__file__, request)],
    )
    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        print(out.fitter)
        print(out.fitter.F_nyquist_Hz)
        print(out.fitter.zeros)
        print(out.fitter.zeros_overlay)
        # print(out.fitter.ZPKrep)
        # assert(False)
        pass
    return


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_L2P_MULT_WOLAY(request, browser, plotsections, plot_verbosity):
    data_name = "L2P_MULT_WLF"
    out = v1.data2filter(
        IIRrational_data_dev(
            data_name,
            instance_num=1,
            set_num=2,
        ),
        delay_s=0,
        ZPK_overlay=((0.99999,), (), 1),
        hints=[sign_validate_and_plot_hint(__file__, request)],
    )
    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        print(out.fitter)
        print(out.fitter.F_nyquist_Hz)
        print(out.fitter.zeros)
        print(out.fitter.zeros_overlay)
        print(out.fitter.ZPKrep)
        # assert(False)
        pass
    return


if __name__ == "__main__":
    test_simple1(False, None, 10)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from wield.bunch.depbunch import (
    DepBunch,
    depB_property,
    NOARG,
)

from .root_bunch import RBAlgorithms
from .zpktf import ZPKTF


class ZPKwData(DepBunch):
    RBalgo = RBAlgorithms()
    root_constraint = RBalgo.root_constraints.mirror_real
    phase_missing = False
    residuals_average = 0

    def __build__(
        self,
        _args=None,
        data=NOARG,
        F_Hz=NOARG,
        W=NOARG,
        delay_s=NOARG,
        zeros=NOARG,
        poles=NOARG,
        gain=NOARG,
        zeros_overlay=NOARG,
        poles_overlay=NOARG,
        ZPK=NOARG,
        ZPK_overlay=NOARG,
        F_nyquist_Hz=NOARG,
        parent=NOARG,
        residuals_log_im_scale=NOARG,
        ZPKrep=NOARG,
    ):

        if _args is not None:
            raise RuntimeError("Only keyword arguments allowed")

        if zeros is NOARG and ZPK is not NOARG:
            zeros = ZPK[0]

        if poles is NOARG and ZPK is not NOARG:
            poles = ZPK[1]

        if zeros_overlay is NOARG and ZPK_overlay is not NOARG:
            zeros_overlay = ZPK_overlay[0]

        if poles_overlay is NOARG and ZPK_overlay is not NOARG:
            poles_overlay = ZPK_overlay[1]

        if gain is NOARG and ZPK is not NOARG:
            gain = ZPK[2]

        if ZPK_overlay is not NOARG:
            gain2 = ZPK_overlay[2]
        else:
            gain2 = 1

        if ZPKrep is not NOARG:
            parent = ZPKrep

        if parent is not NOARG:
            if data is NOARG:
                data = parent.data
            if F_Hz is NOARG:
                F_Hz = parent.F_Hz
            if W is NOARG:
                W = parent.W
            if delay_s is NOARG:
                delay_s = parent.delay_s
            if zeros is NOARG:
                zeros = parent.zeros
            if poles is NOARG:
                poles = parent.poles
            if gain is NOARG:
                gain = parent.gain
            if zeros_overlay is NOARG:
                zeros_overlay = parent.zeros_overlay
            if poles_overlay is NOARG:
                poles_overlay = parent.poles_overlay
            if F_nyquist_Hz is NOARG:
                F_nyquist_Hz = parent.F_nyquist_Hz
            if residuals_log_im_scale is NOARG:
                residuals_log_im_scale = parent.residuals_log_im_scale

        if zeros is NOARG:
            zeros = ()

        if poles is NOARG:
            poles = ()

        if zeros_overlay is NOARG:
            zeros_overlay = ()

        if poles_overlay is NOARG:
            poles_overlay = ()

        if gain is NOARG:
            gain = 1

        if F_Hz is NOARG:
            raise RuntimeError("Must Specify F_Hz")
        if data is NOARG:
            raise RuntimeError("Must Specify data")
        if F_nyquist_Hz is NOARG:
            raise RuntimeError("Must Specify F_nyquist_Hz")
        if W is NOARG:
            if data is None:
                W = None
            else:
                W = 1
        if residuals_log_im_scale is NOARG:
            if data is None:
                residuals_log_im_scale = None
            else:
                residuals_log_im_scale = 1

        if delay_s is NOARG:
            delay_s = 0

        if data is not None:
            if W is None:
                W = 1
            F_Hz, data, W, residuals_log_im_scale = np.broadcast_arrays(
                F_Hz, data, W, residuals_log_im_scale
            )

        self.F_Hz = F_Hz
        self.data = data
        self.W = W
        self.delay_s = delay_s

        self.zeros = zeros
        self.poles = poles
        self.zeros_overlay = zeros_overlay
        self.poles_overlay = poles_overlay
        self.gain = gain * gain2

        self.F_nyquist_Hz = F_nyquist_Hz

        self.residuals_log_im_scale = residuals_log_im_scale

    @depB_property
    def order(self):
        return max(
            len(self.poles) + len(self.poles_overlay),
            len(self.zeros) + len(self.zeros_overlay),
        )

    @depB_property
    def order_sos(self):
        return (
            max(
                len(self.poles) + len(self.poles_overlay),
                len(self.zeros) + len(self.zeros_overlay),
            )
            + 1
        ) // 2

    @depB_property
    def order_relative(self):
        return (len(self.zeros) + len(self.zeros_overlay)) - (
            len(self.poles) + len(self.poles_overlay)
        )

    @depB_property
    def F_nyquist_Hz(self, val):
        if val is None:
            return val
        assert val > 0
        return val

    @depB_property
    def zeros(self, val):
        val = self.RBalgo.expect_atleast(val, constraint=self.root_constraint)
        return val

    @depB_property
    def poles(self, val):
        val = self.RBalgo.expect_atleast(val, constraint=self.root_constraint)
        return val

    @depB_property
    def zeros_overlay(self, val):
        val = self.RBalgo.expect_atleast(val, constraint=self.root_constraint)
        return val

    @depB_property
    def poles_overlay(self, val):
        val = self.RBalgo.expect_atleast(val, constraint=self.root_constraint)
        return val

    @depB_property
    def X_grid(self):
        if self.F_nyquist_Hz is None:
            return 1j * self.F_Hz
        else:
            # use Z^-1
            return np.exp(1j * np.pi * self.F_Hz / self.F_nyquist_Hz)

    @depB_property
    def ZPKrep(self):
        return self

    @depB_property
    def ZPK(self):
        return ZPKTF(
            zeros=self.zeros * self.zeros_overlay,
            poles=self.poles * self.poles_overlay,
            gain=self.gain,
            F_nyquist_Hz=self.F_nyquist_Hz,
        )

    @depB_property
    def ZPKfit(self):
        return ZPKTF(
            (
                self.zeros,
                self.poles,
                self.gain,
            ),
            F_nyquist_Hz=self.F_nyquist_Hz,
        )

    @depB_property
    def ZPKoverlay(self):
        return ZPKTF(
            (
                self.zeros_overlay,
                self.poles_overlay,
                1,
            ),
            F_nyquist_Hz=self.F_nyquist_Hz,
        )

    @depB_property
    def xfer_fit(self):
        # TODO must add pole-zero rephasing for Z filters

        # TODO, make this name consistent in all classes
        h, lnG = self.poles_overlay.val_lnG(self.X_grid)
        h, lnG = self.poles.val_lnG(self.X_grid, h=h, lnG=lnG)

        h, lnG = self.zeros_overlay.val_lnG(self.X_grid, h=1 / h, lnG=-lnG)
        h, lnG = self.zeros.val_lnG(self.X_grid, h=h, lnG=lnG)
        if self.delay_s != 0:
            h = h * np.exp(-2j * np.pi * self.delay_s * self.F_Hz)
        return h * (np.exp(lnG) * self.gain)

    def xfer_eval(self, F_Hz):
        # TODO must add pole-zero rephasing for Z filters

        # TODO, make this name consistent in all classes
        if self.F_nyquist_Hz is None:
            X_grid = 1j * F_Hz
        else:
            # use Z^-1
            X_grid = np.exp(1j * np.pi * F_Hz / self.F_nyquist_Hz)
        h, lnG = self.poles_overlay.val_lnG(X_grid)
        h, lnG = self.poles.val_lnG(X_grid, h=h, lnG=lnG)

        h, lnG = self.zeros_overlay.val_lnG(X_grid, h=1 / h, lnG=-lnG)
        h, lnG = self.zeros.val_lnG(X_grid, h=h, lnG=lnG)
        if self.delay_s != 0:
            h = h * np.exp(-2j * np.pi * self.delay_s * F_Hz)
        return h * (np.exp(lnG) * self.gain)

    @depB_property
    def order(self):
        return max(
            len(self.poles) + len(self.poles_overlay),
            len(self.zeros) + len(self.zeros_overlay),
        )

    @depB_property
    def order_total(self):
        return (
            len(self.poles)
            + len(self.poles_overlay)
            + len(self.zeros)
            + len(self.zeros_overlay)
        )

    @depB_property
    def order_sos(self):
        return (
            max(
                len(self.poles) + len(self.poles_overlay),
                len(self.zeros) + len(self.zeros_overlay),
            )
            + 1
        ) // 2

    @depB_property
    def order_relative(self):
        return (len(self.zeros) + len(self.zeros_overlay)) - (
            len(self.poles) + len(self.poles_overlay)
        )



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

import os
import os.path as path
import contextlib

from wield.iirrational import v1
from wield.iirrational import fitters_rational, fitters_ZPK
from wield.iirrational.testing.plots import plot_on_assert

try:
    from IIRrational_test_data import IIRrational_data_dev
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_quad_M0_P2V_10Hz(request, browser, plotsections, plot_verbosity):
    data_name = "quad_M0_P2V_10Hz"
    out = v1.data2filter(
        IIRrational_data_dev(
            data_name,
            instance_num=1,
            set_num=2,
        ),
        order_initial=80,
    )

    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        pass
        # assert(False)
    return


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_quad_M0_P2V_10Hz_S(request, browser, plotsections, plot_verbosity):
    data_name = "quad_M0_P2V_10Hz"
    out = v1.data2filter(
        IIRrational_data_dev(
            data_name,
            instance_num=1,
            set_num=2,
        ),
        F_nyquist_Hz=None,
        order_initial=80,
        delay_s=0,
    )
    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        pass
        # assert(False)
    return


# @pytest.mark.skipif(module_import_skip, reason = 'cannot import IIRrational_test_data')
# def test_quad_M0_P2V_10Hz_rat_Z(browser, plotsections, plot_verbosity):
#    data_name = 'quad_M0_P2V_10Hz'
#    out = v1.rational_disc_fit(
#        IIRrational_data_dev(
#            data_name,
#            instance_num = 1,
#            set_num = 2,
#        ),
#        nyquist_final_Hz = 1e3,
#        order = 100,
#    )
#
#    import matplotlib.pyplot as plt
#    from wield.iirrational import plots
#    axB = plots.plot_fitter_flag(
#        out,
#        #don't use jpg as matplotlib can't write it out of the box (needs pillow)
#        fname = path.join(path.split(__file__)[0], 'out-cplx/test-rational-Z.png'),
#    )
#    plt.close(axB.fig)
#    plt.show()
#    return
#
# @pytest.mark.skipif(module_import_skip, reason = 'cannot import IIRrational_test_data')
# def test_quad_M0_P2V_10Hz_rat_S(browser, plotsections, plot_verbosity):
#    data_name = 'quad_M0_P2V_10Hz'
#    out = v1.rational_disc_fit(
#        IIRrational_data_dev(
#            data_name,
#            instance_num = 1,
#            set_num = 2,
#        ),
#        nyquist_final_Hz = 0,
#        order = 100,
#    )
#
#    import matplotlib.pyplot as plt
#    from wield.iirrational import plots
#    axB = plots.plot_fitter_flag(
#        out,
#        #don't use jpg as matplotlib can't write it out of the box (needs pillow)
#        fname = path.join(path.split(__file__)[0], 'out-cplx/test-rational-S.png'),
#    )
#    plt.close(axB.fig)
#    plt.show()
#    return


if __name__ == "__main__":
    test_simple1(False, None, 10)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np

from ..utilities import ensure_aid

# from .. import fitters_ZPK


def sign_check_flip(fitter):
    """ """
    xfer = fitter.xfer_fit
    data = fitter.data
    rat = data / xfer
    rat_ang = np.exp(1j * np.angle(rat))
    ang_avg_fit = np.sum(rat_ang * fitter.W ** 2) / np.sum(fitter.W ** 2)

    if ang_avg_fit.real < 0:
        fitter.gain = -fitter.gain


def flip_mindelay_opt(aid):
    """
    Attempts to flip each non-mindelay zero and then reoptimize
    """
    aid = ensure_aid(aid)

    # TODO, also deal with real roots
    min_idx = 0

    while True:
        coding_lst = list(aid.fitter.num_codings)
        for idx, coding in enumerate(aid.fitter.num_codings):
            if idx <= min_idx:
                # TODO, change logic to not need this and be faster
                continue
            zeros = coding.roots_c_Sf()
            if zeros:
                z = zeros[0]
                if z.real > 0:
                    aid.log("trying flipping", z)
                    # using a coding which cannot flip over,
                    # since that affects the sign of the gain
                    coding_ins = aid.fitter.coding_map.num_c(aid.fitter)
                    # flip the root over, but also reduce its effect to help convergence
                    coding_ins.update_roots_Sf((-z).conjugate())
                    coding_lst[idx] = coding_ins
                    min_idx = idx
                    break
        else:
            # breaks from the while loop only if for-loop break doesn't occur
            break

        fitter_new = aid.fitter.__class__(
            parent=aid.fitter,
            num_codings=coding_lst,
        )
        # TODO, note, should this only be flipped in s-domain?
        # flip the gain,
        # print("GAIN: ", aid.fitter.xfer_fit / fitter_new.xfer_fit)
        # fitter_new.gain = -fitter_new.gain
        # print("GAIN: ", aid.fitter.xfer_fit / fitter_new.xfer_fit)

        with fitter_new.with_codings_only([coding_ins]):
            fitter_new.optimize(aid=aid)
        fitter_new.optimize(aid=aid)
        # print("GAIN3: ", aid.fitter.xfer_fit / fitter_new.xfer_fit)

        aid.fitter_check(
            fitter_new,
            hint_name="mindelay_opt",
            variant="OrdC",
        )
    return


def insert_triplets(aid):
    """
    iserts on each poles and zeros a complex and real roots with bandwidth 2x the data
    """
    aid = ensure_aid(aid)

    cplx_t = aid.fitter.coding_map.num_c
    real_t = aid.fitter.coding_map.num_r

    F_l_Hz = aid.fitter.F_max_Hz
    BW_2x_Hz = 2 * F_l_Hz
    F_high_Hz = 0.90 * F_l_Hz

    coding_num_p = cplx_t(aid.fitter)
    coding_num_p.update_roots_Sf(-BW_2x_Hz + 1j * F_high_Hz)
    coding_num_p2 = real_t(aid.fitter)
    coding_num_p2.update_roots_Sf(-BW_2x_Hz)
    coding_den_p2 = real_t(aid.fitter)
    coding_den_p2.update_roots_Sf(-BW_2x_Hz)
    coding_den_p = cplx_t(aid.fitter)
    coding_den_p.update_roots_Sf(-BW_2x_Hz + 1j * F_high_Hz)

    fitter_new = aid.fitter.__class__(
        parent=aid.fitter,
        num_codings=aid.fitter.num_codings + [coding_num_p2],
        den_codings=aid.fitter.den_codings + [coding_den_p2],
    )
    res_pre = fitter_new.residuals_average
    with fitter_new.with_codings_only(
        [coding_num_p, coding_num_p2] + [coding_den_p, coding_den_p2]
    ):
        fitter_new.optimize(aid=aid)
    res_mid = fitter_new.residuals_average
    fitter_new.optimize(aid=aid)
    res_post = fitter_new.residuals_average

    res_ratio = fitter_new.residuals_average / aid.fitter.residuals_average

    aid.log(
        "TRIPLETS (rat = {0}, pre = {1}, mid = {2}, post = {3})".format(
            res_ratio, res_pre, res_mid, res_post
        )
    )
    return aid.fitter_check(
        fitter_new,
        hint_name="insert_triplets2",
        variant="OrdUp",
    )


def set_min_BW(aid):
    aid = ensure_aid(aid)

    def modify_codings(codings):
        for coding in codings:
            roots = coding.roots_c_Sf()
            if roots:
                (r,) = roots
                root_F_Hz = r.imag
                F_idx = np.searchsorted(aid.fitter.F_Hz, root_F_Hz)
                F_idx_low = max(0, F_idx - 2)
                F_idx_high = min(len(aid.fitter.F_Hz) - 1, F_idx + 2)
                # get the local mean width of bins to restrict bandwidths
                F_mean_Hz = (
                    aid.fitter.F_Hz[F_idx_high] - aid.fitter.F_Hz[F_idx_low]
                ) / (F_idx_high - F_idx_low)
                coding.option_set(minimum_BW_Hz=F_mean_Hz * 0.5)

    modify_codings(aid.fitter.den_codings)
    modify_codings(aid.fitter.num_codings)

    aid.fitter.distance_limit_auto = 2
    aid.fitter_update()
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from .base import (
    CodingType,
    Ipi,
    # I2pi
)

# import scipy.linalg


class CodingRealAmp(CodingType):
    N_parameters = 1
    unstable = False
    p_amplitude = 0
    lock_amplitude = False
    hide_amplitude = False
    restabilize = False

    def setup(
        self,
        hide_all=None,
        lock_amplitude=None,
        hide_amplitude=None,
        disable=None,
    ):
        if hide_all:
            hide_amplitude = True

        if lock_amplitude is not None:
            self.lock_amplitude = lock_amplitude
        if hide_amplitude is not None:
            self.hide_amplitude = hide_amplitude

        if self.hide_amplitude:
            N_parameters = 0
        else:
            N_parameters = 1

        if disable is not None:
            self.disable = disable
        if self.disable:
            N_parameters = 0

        self.N_parameters = N_parameters

    def update(self, amplitude=None):
        if self.disable:
            assert amplitude is None
            return
        if self.hide_amplitude:
            assert amplitude is None
        else:
            if not self.lock_amplitude:
                self.p_amplitude = amplitude

    def reduce(self):
        if self.disable:
            return []
        if self.hide_amplitude:
            return []
        else:
            return [self.p_amplitude]

    def update_roots(self, r1):
        assert r1.imag == 0
        if self.restabilize:
            if r1 > 1:
                r1 = 1 / r1
        self.p_amplitude = r1
        return

    @property
    def F_Hz(self):
        return 0

    @property
    def amplitude(self):
        return self.p_amplitude

    def transfer(self):
        # real, linear amplitude
        return 1 - self.p_amplitude * sys.Xn_grid

    def derivative(self):
        if self.disable:
            return []
        # real/imaginary part of root
        if not self.hide_amplitude:
            jac = [
                -sys.Xn_grid if not self.lock_amplitude else 0,
            ]
            return jac
        else:
            return []

    def roots_r(self):
        return [self.p_amplitude]


class CodingRealAmpUX(CodingRealAmp):
    restabilize = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
import scipy
import scipy.linalg
import scipy.signal

from wield import declarative
from wield.bunch import DeepBunch
from wield.bunch.hdf_deep_bunch import HDFDeepBunch

from ..plots import (
    plot_fit,
    plot_ZP,
    plot_fitter_flag,
    plot_fitter_flag_compare,
    plot_ZP_grab,
)

from ..utilities.ipynb.displays import *

from ..utilities.np import logspaced
from ..utilities.mpl import mplfigB, generate_stacked_plot_ax

from ..fitters_ZPK import ZPKrep2MRF, MRF
from ..fitters_rational import RationalDiscFilter, ChebychevFilter

from .. import v1
from .. import v2

from ..testing import IIRrational_data


# run version printer from function to not further pollute namespace
def print_version():
    from .. import auto_version
    from .. import version

    print(
        "wield.iirrational version: {} (git:{})".format(
            version, auto_version.git_shorthash
        )
    )


print_version()
del print_version



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from .base import (
    CodingType,
)

# import scipy.linalg


class CodingRealnlA(CodingType):
    N_parameters = 1
    unstable = False
    p_nl_A = 0
    lock_nl_A = None
    hide_nl_A = None

    def setup(
        self,
        hide_all=None,
        lock_nl_A=None,
        hide_nl_A=None,
        hide_amplitude=None,
        disable=None,
    ):
        if hide_all:
            hide_nl_A = True
        if hide_amplitude is not None:
            self.hide_nl_A = hide_amplitude

        if lock_nl_A is not None:
            self.lock_nl_A = lock_nl_A
        if hide_nl_A is not None:
            self.hide_nl_A = hide_nl_A
        if self.hide_nl_A:
            N_parameters = 0
        else:
            N_parameters = 1

        if disable is not None:
            self.disable = disable
        if self.disable:
            N_parameters = 0

        self.N_parameters = N_parameters

    def update(self, nl_A=None):
        if self.disable:
            assert nl_A is None
            return
        if self.hide_nl_A:
            assert nl_A is None
        else:
            if not self.lock_nl_A:
                self.p_nl_A = nl_A

    def reduce(self):
        if self.disable:
            return []
        if self.hide_nl_A:
            return []
        else:
            return [self.p_nl_A]

    def update_roots(self, r1):
        assert r1.imag == 0
        amp = r1

        if abs(amp) < 1:
            # amp = x / (1. + x**2)**.5
            # amp**2 = x**2 / (1. + x**2)
            # amp**2 (1. + x**2) = x**2
            # amp**2  = x**2 ( 1 - amp**2)
            # amp**2 / (1 - amp**2) = x**2
            self.unstable = False
            self.p_nl_A = amp / (1 - amp ** 2) ** 0.5
        else:
            # amp = (1. + x**2)**.5 / x
            # amp**2 =  (1. + x**2) / x**2
            # amp**2 x**2 = x**2 + 1
            # amp**2  = x**2 (amp**2 - 1)
            # amp**2 / (amp**2 - 1) = x**2
            self.unstable = True
            self.p_nl_A = amp / (amp ** 2 - 1) ** 0.5
        return
        return

    @property
    def amplitude(self):
        if not self.unstable:
            amp = self.p_nl_A / (1.0 + self.p_nl_A ** 2) ** 0.5
        else:
            amp = (1.0 + self.p_nl_A ** 2) ** 0.5 / self.p_nl_A
        return amp

    @property
    def F_Hz(self):
        return 0

    def transfer(self):
        # real, log amplitude, positive
        return 1 - self.amplitude * sys.Xn_grid

    def derivative(self):
        if self.disable:
            return []
        # real/imaginary part of root
        if not self.unstable:
            sqp5 = (1.0 + self.p_nl_A ** 2) ** 0.5
            # amp = self.p_nl_A / sqp5
            DampDl = 1 / sqp5 - self.p_nl_A ** 2 / (1.0 + self.p_nl_A ** 2) ** 1.5
        else:
            sqp5 = (1.0 + self.p_nl_A ** 2) ** 0.5
            # amp = sqp5 / self.p_nl_A
            DampDl = 1 / sqp5 - sqp5 / self.p_nl_A ** 2

        if not self.hide_nl_A:
            return [
                -DampDl * sys.Xn_grid if not self.lock_nl_A else 0,
            ]
        else:
            return []

    def roots_r(self):
        return [self.amplitude]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from matplotlib import gridspec

from ..utilities.np import logspaced
from ..utilities.mpl import generate_stacked_plot_ax


def plot_rel_comparison(
    self,
    fitter,
    fitter_ref,
    N_points=1000,
    gs_base=gridspec.GridSpec(1, 1)[0],
    fig=None,
    ax_bunch=None,
    xscale=None,
):
    if ax_bunch is None:
        axB = generate_stacked_plot_ax(
            [
                ("rel", True),
            ],
            height_ratios=dict(
                mag=1,
            ),
            heights_phys_in_default=1,
            xscales="log",
            gs_base=gs_base,
            fig=fig,
            ax_bunch=ax_bunch,
        )
    else:
        axB = ax_bunch

    axB.phase = axB.rel.twinx()
    f_min = np.max(fitter.F_Hz)
    f = np.sort(
        np.concatenate(
            [
                fitter.F_Hz,
                np.linspace(f_min, fitter.F_nyquist_Hz, N_points // 2),
                logspaced(f_min, fitter.F_nyquist_Hz, N_points // 4),
            ]
        )
    )
    f = f[f > 0]
    h = fitter.xfer_eval(f)
    h_ref = fitter_ref.xfer_eval(f)

    rel = h / h_ref

    mline = axB.rel.semilogx(
        f, 20 * np.log10(abs(rel)), color="black", label="rel mag [db]"
    )
    pline = axB.phase.semilogx(
        f,
        np.angle(rel),
        label="fit phase [rad]",
        color="purple",
    )

    axB.rel.legend(handles=[mline[0], pline[0]], ncol=2, fontsize=8, loc="lower right")
    axB.rel.set_xlabel("Frequency [Hz]")
    axB.rel.set_ylabel("DB difference")
    axB.phase.set_ylabel("Rad Difference")

    axB.finalize()
    return axB



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from .base import CodingType, I2pi


class CodingDelayNL(CodingType):
    N_parameters = 1
    p_delay_nl = 0
    delay_s_min = 0
    delay_s_diff = 0
    deriv_deadzone = 1e-2

    def __init__(self, sys):
        super(CodingDelayNL, self).__init__(sys)

    def clone(self, sys):
        cpy = super(CodingDelayNL, self).clone(sys)
        cpy.delay_s = self.delay_s
        return cpy

    def setup(self, delay_s=None):
        if delay_s is not None:
            self.delay_s = delay_s
        return

    def update(self, delay_nl=None):
        if self.N_parameters > 0:
            self.p_delay_nl = delay_nl
        else:
            assert delay_nl is None

    def reduce(self):
        if self.N_parameters > 0:
            return [self.p_delay_nl]
        else:
            return []

    @property
    def N_parameters(self):
        self.delay_s_min = self.sys.delay_s_min
        self.delay_s_diff = self.sys.delay_s_max - self.sys.delay_s_min

        if self.delay_s_diff <= 0:
            N_parameters = 0
            self.delay_s_diff = 0
        else:
            N_parameters = 1
        return N_parameters

    @property
    def delay_s(self):
        return self.delay_s_min + self.delay_s_diff * (1 - np.cos(self.p_delay_nl)) / 2

    @delay_s.setter
    def delay_s(self, val):
        if self.delay_s_diff <= 0:
            self.p_delay_nl = 0
            return
        if val > self.delay_s_min:
            diffrat = (val - self.delay_s_min) / self.delay_s_diff
            if diffrat > 1:
                self.p_delay_nl = np.pi
            else:
                self.p_delay_nl = np.arccos(1 - diffrat * 2)
        else:
            self.p_delay_nl = 0

    def transfer(self):
        return np.exp(-I2pi * self.delay_s * self.sys.F_Hz)

    @property
    def derivative_deadzoned(self):
        sD = np.sin(self.p_delay_nl)
        if abs(sD) < self.deriv_deadzone:
            return True
        return False

    def derivative(self):
        if self.N_parameters == 0:
            return []
        sD = np.sin(self.p_delay_nl)
        if sD > 0:
            if sD < self.deriv_deadzone:
                sD = self.deriv_deadzone
        else:
            if sD > -self.deriv_deadzone:
                sD = -self.deriv_deadzone
        dDdP = self.delay_s_diff * sD / 2
        return [
            -I2pi * self.sys.F_Hz * dDdP * np.exp(-I2pi * self.delay_s * self.sys.F_Hz)
        ]

    def roots(self):
        return []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from .base import (
    EmptyCopy,
    CodingType,
    CodingTypeZ,
    BranchCutAmbiguity,
    Ipi,
    I2pi,
)

from .delay_nl import (
    CodingDelayNL,
)

from .delay_pair import (
    CodingDelayPairNl,
    CodingDelayPair,
)

from .gain_delay import (
    CodingGain,
    CodingDelay,
)



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

from wield.iirrational import v2
from wield.iirrational.testing.plots import plot_on_assert, digest_on_assert

from wield.iirrational.testing.utilities import (
    sign_validate_and_plot_hint,
)

try:
    from IIRrational_test_data import IIRrational_data_dev
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_HTTS_P_L(request, browser, plotsections, plot_verbosity):
    data = IIRrational_data_dev("HTTS_P_L")
    out = v2.data2filter(
        F_Hz=data.F_Hz,
        data=data.data,
        SNR=data.SNR,
        order_initial=20,
        hints=[sign_validate_and_plot_hint(__file__, request)],
    )
    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        pass
    # with digest_on_assert(__file__, request, out, plot_anyway = True):
    #    pass
    return


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_HTTS_Y_L(request, browser, plotsections, plot_verbosity):
    data = IIRrational_data_dev("HTTS_Y_L")
    out = v2.data2filter(
        F_Hz=data.F_Hz,
        data=data.data,
        SNR=data.SNR,
        order_initial=20,
        hints=[sign_validate_and_plot_hint(__file__, request)],
    )
    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        pass
    return


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_HTTS_Y_L_reldeg(request, browser, plotsections, plot_verbosity):
    def reldeg_validate(aid, fitter):
        rdmax = aid.hint("relative_degree_max")
        rdmin = aid.hint("relative_degree_min")
        orders = aid.fitter_orders(fitter)
        # print("DEGREE", orders.total, orders.reldeg, rdmin, rdmax)
        if rdmax is not None:
            assert orders.reldeg <= rdmax
        if rdmin is not None:
            assert orders.reldeg >= rdmin

    hint = {
        "fitter_update_validate": reldeg_validate,
        "fitter_check_validate": reldeg_validate,
    }
    data = IIRrational_data_dev("HTTS_Y_L")
    out = v2.data2filter(
        F_Hz=data.F_Hz,
        data=data.data,
        SNR=data.SNR,
        order_initial=20,
        hints=[hint],
        relative_degree=-2,
        log_level=10,
        log_level_debug=10,
    )
    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        pass
    return



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest
import os
import numpy as np
import os.path as path

from wield.iirrational.testing import IIRrational_data
from wield.iirrational.utilities.print import pprint
from wield.iirrational.v2.__main__ import main as v2_main

try:
    from IIRrational_test_data import matlab as matlab_test_data
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False

localpath = path.split(__file__)[0]
skipdeco = pytest.mark.skipif(
    module_import_skip, reason="cannot import IIRrational_test_data"
)


def gen_outname(base):
    fname_out = path.join(localpath, base)
    try:
        os.unlink(fname_out)
    except OSError:
        pass
    return fname_out


@skipdeco
def test_mat_datamissing():
    fname = matlab_test_data.matfiles["iir_test_struct.mat"]
    fname_out = path.join(localpath, "out.mat")
    with pytest.raises(SystemExit):
        v2_main(["-D", "test,D", "--mode", "fit", fname, fname_out])


@skipdeco
def test_mat_help():
    fname = matlab_test_data.matfiles["iir_test_struct.mat"]
    with pytest.raises(SystemExit):
        v2_main(["-h"])


@skipdeco
def test_mat_printdata():
    fname = matlab_test_data.matfiles["iir_test_struct.mat"]
    fname_out = path.join(localpath, "out.mat")
    with pytest.raises(SystemExit):
        v2_main(["-D", "data,D", "--mode", "printdata", fname, fname_out])


# TODO test printsettings
# TODO test printconfig


@skipdeco
def test_mat_fit():
    fname = matlab_test_data.matfiles["iir_test_struct.mat"]
    fname_out = gen_outname("out.json")
    v2_main(
        [
            "-D",
            "data,D",
            "--mode",
            "fit",
            fname,
            fname_out,
            "--choose",
            "baseline",
            "-p=-2",
        ]
    )

    fname_out2 = gen_outname("out2.yaml")
    v2_main(
        ["-D", "data,D", "--mode", "fit", fname_out, fname_out2, "--choose", "baseline"]
    )

    fname_out3 = gen_outname("out3.mat")
    v2_main(
        [
            "-D",
            "data,D",
            "--mode",
            "fit",
            fname_out2,
            fname_out3,
            "--choose",
            "baseline",
        ]
    )

    fname_out4 = gen_outname("out4.h5")
    v2_main(
        [
            "-D",
            "data,D",
            "--mode",
            "fit",
            fname_out3,
            fname_out4,
            "--choose",
            "baseline",
        ]
    )

    fname_out5 = gen_outname("out5.pkl")
    v2_main(
        [
            "-D",
            "data,D",
            "--mode",
            "fit",
            fname_out4,
            fname_out5,
            "--choose",
            "baseline",
        ]
    )


@skipdeco
def test_csv_fit():
    fname = matlab_test_data.matfiles["iir_test_struct.mat"]
    from wield.iirrational.file_io import load_any

    fdict = load_any(fname=fname, ftype="mat")
    data = fdict["data"]
    darr = np.vstack(
        [
            data["F_Hz"],
            data["xfer"].real,
            data["xfer"].imag,
            data["SNR"],
        ]
    )
    fname_in = path.join(localpath, "in.csv")
    np.savetxt(fname_in, darr, delimiter=",")
    fname_out = gen_outname("out.mat")

    v2_main(
        [
            "-D",
            "data,D",
            "--mode",
            "fit",
            fname_in,
            fname_out,
            "--choose",
            "baseline",
            "-p=-30",
        ]
    )


@skipdeco
def test_mat_full():
    fname = matlab_test_data.matfiles["iir_test_struct.mat"]
    fname_out = gen_outname("out.mat")
    v2_main(
        [
            fname,
            fname_out,
            "--choose",
            "baseline",
            "--relative_degree_min",
            "-2",
        ]
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from .base import (
    CodingType,
    Ipi,
    # I2pi
)

# import scipy.linalg


class CodingFnlA(CodingType):
    N_parameters = 2
    unstable = False
    p_F_Hz = 0
    p_nl_BW = 0
    lock_F_Hz = False
    lock_nl_A = False
    hide_F_Hz = False
    hide_nl_A = False

    def setup(
        self,
        hide_all=None,
        lock_F_Hz=None,
        hide_F_Hz=None,
        lock_nl_A=None,
        hide_nl_A=None,
        hide_amplitude=None,
        disable=None,
    ):
        if hide_all:
            hide_F_Hz = True
            hide_nl_A = True

        if hide_amplitude is not None:
            self.hide_nl_A = hide_amplitude

        N_parameters = 2
        if lock_F_Hz is not None:
            self.lock_F_Hz = lock_F_Hz
        if hide_F_Hz is not None:
            self.hide_F_Hz = hide_F_Hz
        if self.hide_F_Hz:
            N_parameters -= 1

        if lock_nl_A is not None:
            self.lock_nl_A = lock_nl_A
        if hide_nl_A is not None:
            self.hide_nl_A = hide_nl_A
        if self.hide_nl_A:
            N_parameters -= 1

        if disable is not None:
            self.disable = disable
        if self.disable:
            N_parameters = 0

        self.N_parameters = N_parameters

    def update(self, A=None, B=None):
        if self.disable:
            assert A is None and B is None
            return
        if self.hide_nl_A:
            assert B is None
            if self.hide_F_Hz:
                assert A is None
            else:
                F_Hz = A
                if not self.lock_F_Hz:
                    self.p_F_Hz = F_Hz
        else:
            if self.hide_F_Hz:

                nl_A = A
                if not self.lock_nl_A:
                    self.p_nl_BW = nl_A
            else:
                F_Hz, nl_A = A, B
                if not self.lock_F_Hz:
                    self.p_F_Hz = F_Hz

                if not self.lock_nl_A:
                    self.p_nl_BW = nl_A

    def reduce(self):
        if self.disable:
            return []
        if self.hide_nl_A:
            if self.hide_F_Hz:
                return []
            else:
                return [self.p_F_Hz]
        else:
            if self.hide_F_Hz:
                return [self.p_nl_BW]
            else:
                return [
                    self.p_F_Hz,
                    self.p_nl_BW,
                ]

    def update_roots(self, r1):
        """ """
        F_Hz = r1.imag
        BW_Hz = r1.real
        self.p_F_Hz = F_Hz

        if BW_Hz < 0:
            # amp = x / (1. + x**2)**.5
            # amp**2 = x**2 / (1. + x**2)
            # amp**2 (1. + x**2) = x**2
            # amp**2  = x**2 ( 1 - amp**2)
            # amp**2 / (1 - amp**2) = x**2
            self.unstable = False
            self.p_nl_BW = amp / (1 - amp ** 2) ** 0.5
        else:
            # amp = (1. + x**2)**.5 / x
            # amp**2 =  (1. + x**2) / x**2
            # amp**2 x**2 = x**2 + 1
            # amp**2  = x**2 (amp**2 - 1)
            # amp**2 / (amp**2 - 1) = x**2
            self.unstable = True
            self.p_nl_BW = amp / (amp ** 2 - 1) ** 0.5
        return

    @property
    def F_Hz(self):
        return self.p_F_Hz

    @property
    def BW_Hz(self):
        if not self.unstable:
            amp = self.p_nl_BW / (1.0 + self.p_nl_BW ** 2) ** 0.5
        else:
            amp = (1.0 + self.p_nl_BW ** 2) ** 0.5 / self.p_nl_BW
        return amp

    def transfer(self):
        # frequency, logarithmic BW_Hz
        BW = self.BW_Hz
        F = self.F_Hz
        X = sys.Xn_grid
        Xsq = sys.Xn_grid_sq
        return (r * r + i * i) * Xsq - 2 * Xn * r + 1

    def derivative(self):
        if self.disable:
            return []
        # real/imaginary part of root
        if not self.unstable:
            sqp5 = (1.0 + self.p_nl_BW ** 2) ** 0.5
            amp = self.p_nl_BW / sqp5
            DampDl = 1 / sqp5 - self.p_nl_BW ** 2 / (1.0 + self.p_nl_BW ** 2) ** 1.5
        else:
            sqp5 = (1.0 + self.p_nl_BW ** 2) ** 0.5
            amp = sqp5 / self.p_nl_BW
            DampDl = 1 / sqp5 - sqp5 / self.p_nl_BW ** 2

        F_nyquist_Hz = sys.F_nyquist_Hz
        X = sys.Xn_grid
        Xsq = sys.Xn_grid_sq
        if not self.hide_F_Hz:
            if not self.hide_nl_A:
                rcos = np.cos(np.pi * self.p_F_Hz / F_nyquist_Hz)
                rsin = np.sin(np.pi * self.p_F_Hz / F_nyquist_Hz)
                return [
                    (2 * amp * np.pi / sys.F_nyquist_Hz) * (X * rsin)
                    if not self.lock_F_Hz
                    else 0,
                    ((2 * amp * DampDl) * Xsq - Xn * (2 * rcos * DampDl))
                    if not self.lock_nl_A
                    else 0,
                ]
            else:
                rsin = np.sin(np.pi * self.p_F_Hz / F_nyquist_Hz)
                return [
                    (2 * amp * np.pi / sys.F_nyquist_Hz) * (X * rsin)
                    if not self.lock_F_Hz
                    else 0,
                ]
        else:
            if not self.hide_nl_A:
                rcos = np.cos(np.pi * self.p_F_Hz / F_nyquist_Hz)
                return [
                    ((2 * amp * DampDl) * Xsq - Xn * (2 * rcos * DampDl))
                    if not self.lock_nl_A
                    else 0,
                ]
            else:
                return []

    def roots_c(self):
        # real/imaginary part of root
        return [self.BW_Hz + 1j * self.F_Hz]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np

from ..codings_cmn import (
    CodingType,
    # Ipi,
    # I2pi
)


class CodingSOSnl(CodingType):
    N_parameters = 2
    p_nl_c1 = 0
    p_nl_c2 = 0
    deriv_deadzone = 1e-4

    # TODO
    max_BW_Hz = 1e4
    max_BW_HzSq = 1e8

    # should be a positive number
    min_BW_Hz = 0

    # True false means mirror to respective side
    unstable = False
    single_root = False

    def update(self, nl_c1, nl_c2=None):
        if self.single_root:
            self.p_nl_c1 = nl_c1
        else:
            assert nl_c2 is not None
            self.p_nl_c1 = nl_c1
            self.p_nl_c2 = nl_c2

    def reduce(self):
        if self.single_root:
            return [self.p_nl_c1]
        else:
            return [self.p_nl_c1, self.p_nl_c2]

    @property
    def F_Hz(self):
        if self.single_root:
            return 0

        c1 = self.p_nl_c1 ** 2
        c1 = c1 / (c1 / self.max_BW_Hz + 1)
        c2 = self.p_nl_c2 ** 2
        c2 = c2 / (c2 / self.max_BW_HzSq + 1)
        disc = c1 * c1 - 4 * c2
        if disc > 0:
            return 0
        else:
            return ((-disc) ** 0.5) / 2

    @property
    def gain_effect(self):
        rs = self.roots_r()
        if len(rs) == 0:
            return 1
        elif len(rs) == 1:
            return 1 if rs[0] < 0 else -1
        else:
            if rs[0] < 0:
                return 1 if rs[1] < 0 else -1
            else:
                return 1 if rs[1] > 0 else -1

    def check_dist_limit(self, F_Hz=None, thresh=1):
        if self.sys.distance_limit_auto >= thresh:
            self.max_BW_Hz = self.sys.max_BW_Hz
            self.max_BW_HzSq = self.max_BW_Hz ** 2
            if F_Hz is None:
                self.min_BW_Hz, D = self.sys.distance_limit(
                    self.F_Hz, with_derivative=True
                )
                # assert(self.min_BW_Hz > 0)
            else:
                self.min_BW_Hz, D = self.sys.distance_limit(F_Hz, with_derivative=True)
                # assert(self.min_BW_Hz > 0)
            return D
        return 0

    def transfer(self):
        if self.single_root:
            self.check_dist_limit(F_Hz=0, thresh=2)
            # second order sections (2x roots either real or complex conj)
            c1 = self.p_nl_c1 ** 2
            c1 = c1 / (c1 / self.max_BW_Hz + 1)
            X = self.sys.Xsf_grid
            if not self.unstable:
                return (c1 + self.min_BW_Hz) + X
            else:
                return -(c1 + self.min_BW_Hz) + X
        else:
            # second order sections (2x roots either real or complex conj)
            c1 = self.p_nl_c1 ** 2
            c2 = self.p_nl_c2 ** 2
            c1 = c1 / (c1 / self.max_BW_Hz + 1)
            c2 = c2 / (c2 / self.max_BW_HzSq + 1)
            disc = c1 * c1 - 4 * c2
            if disc > 0:
                self.check_dist_limit(F_Hz=0, thresh=2)
            else:
                F_Hz = ((-disc) ** 0.5) / 2
                self.check_dist_limit(F_Hz=F_Hz, thresh=2)
            V = self.min_BW_Hz
            X = self.sys.Xsf_grid
            Xsq = self.sys.Xsf_grid_sq
            # V = 0
            if self.unstable:
                xfer = (c2 + V * (V + c1)) - (X * (c1 + 2 * V) - Xsq)
            else:
                xfer = (c2 + V * (V + c1)) + (X * (c1 + 2 * V) + Xsq)
            return xfer

    @property
    def derivative_deadzoned(self):
        if abs(self.p_nl_c1) < self.deriv_deadzone:
            return True
        if not self.single_root:
            if abs(self.p_nl_c2) < self.deriv_deadzone:
                return True
        return False

    def derivative(self):
        c1 = self.p_nl_c1 ** 2
        D1 = c1 / self.max_BW_Hz + 1
        pD_c1 = 2 * self.p_nl_c1 / D1 * (1 - c1 / self.max_BW_Hz)
        if self.p_nl_c1 > 0:
            if self.p_nl_c1 < self.deriv_deadzone:
                pD_c1 = 2 * self.deriv_deadzone
        else:
            if self.p_nl_c1 > -self.deriv_deadzone:
                pD_c1 = -2 * self.deriv_deadzone

        if self.single_root:
            self.check_dist_limit(F_Hz=0, thresh=2)

            if not self.unstable:
                return [pD_c1]
            else:
                return [-pD_c1]
        else:
            c1 = c1 / D1
            c2 = self.p_nl_c2 ** 2
            D2 = c2 / self.max_BW_HzSq + 1
            c2 = c2 / D2

            pD_c2 = 2 * self.p_nl_c2 / D2 * (1 - c2 / self.max_BW_HzSq)
            if self.p_nl_c2 > 0:
                if self.p_nl_c2 < self.deriv_deadzone:
                    pD_c2 = 2 * self.deriv_deadzone
            else:
                if self.p_nl_c2 > -self.deriv_deadzone:
                    pD_c2 = -2 * self.deriv_deadzone

            disc = c1 * c1 - 4 * c2
            if disc > 0:
                V_D = self.check_dist_limit(F_Hz=0, thresh=2)
                V_D_c1 = 0
                V_D_c2 = 0
            else:
                F_Hz = ((-disc) ** 0.5) / 2
                V_D = self.check_dist_limit(F_Hz=F_Hz, thresh=2)
                V_D_c1 = -pD_c1 * c1 / disc * V_D
                V_D_c2 = -pD_c2 / (2 * disc) * V_D

            # TODO, the derivative inclusion of V_D_c1,2 isn't tested well, but
            # that is partially because it doesn't matter much
            V_D_c1 = 0
            V_D_c2 = 0

            V = self.min_BW_Hz

            X = self.sys.Xsf_grid
            if not self.unstable:
                return [
                    (pD_c1 + 2 * V_D_c1) * X + 2 * V * (pD_c1 + V_D_c1),
                    (2 * X * V_D_c2) + pD_c2 + V_D_c2 * (2 * V + c1),
                ]
            else:
                return [
                    -(pD_c1 + 2 * V_D_c1) * X + 2 * V * (pD_c1 + V_D_c1) + pD_c1 * V,
                    -(2 * X * V_D_c2) + pD_c2 + V_D_c2 * (2 * V + c1),
                ]

    def update_roots(self, r1, r2=None):
        """
        r2, may be unspecified, in which case it is assumed to be nothing, if r1 is real, or otherwise the conjugate of r1
        """
        # TODO, incorporate effects of max_BW_Hz
        if r2 is None and r1.imag == 0:
            self.check_dist_limit(F_Hz=0, thresh=1)
            self.single_root = True
            self.N_parameters = 1
            # TODO, fix for S domain
            if self.unstable is not None:
                if r1.real > 0:
                    self.unstable = True
                    r1p = r1.real
                else:
                    self.unstable = False
                    r1p = -r1.real
                if r1p > self.min_BW_Hz:
                    c1 = r1p - self.min_BW_Hz
                    if c1 >= self.max_BW_Hz:
                        c1 = self.max_BW_Hz
                        ret = False
                    else:
                        c1 = c1 / (1 - c1 / self.max_BW_Hz)
                        ret = True
                    self.p_nl_c1 = c1**0.5
                    self.p_nl_c2 = 0
                    return ret
                else:
                    self.p_nl_c1 = 0
                    self.p_nl_c2 = 0
                    return False
        else:
            self.single_root = False
            self.N_parameters = 2
            if r1.real > 0:
                self.unstable = True
                r1p = r1.real
            else:
                self.unstable = False
                r1p = -r1.real

            if r1.imag != 0:
                F_Hz = abs(r1.imag)
                self.check_dist_limit(F_Hz=F_Hz, thresh=1)
                if r1p > self.min_BW_Hz:
                    ret = True
                    r1p = r1p - self.min_BW_Hz
                else:
                    ret = False
                    r1p = 0
                # TODO check conjugates

                c2 = r1p**2 + r1.imag**2
                c1 = 2 * r1p
            else:
                self.check_dist_limit(F_Hz=0, thresh=1)
                if r2 is None:
                    r2 = r1.conjugate()
                if r1p > self.min_BW_Hz:
                    r1p = r1p - self.min_BW_Hz
                    ret = True
                else:
                    ret = False
                    r1p = 0

                if self.unstable:
                    if r2 < 0:
                        raise RuntimeError(
                            "Can't share stable/unstable roots in this coding"
                        )
                    r2p = r2.real
                else:
                    if r2 > 0:
                        raise RuntimeError(
                            "Can't share stable/unstable roots in this coding"
                        )
                    r2p = -r2.real

                if r2p > self.min_BW_Hz:
                    r2p = r2p - self.min_BW_Hz
                else:
                    ret = False
                    r2p = 0

                c2 = r1p * r2p
                c1 = r1p + r2p

            if c2 >= self.max_BW_HzSq:
                ret = False
                c2 = self.max_BW_HzSq
            else:
                c2 = c2 / (1 - c2 / self.max_BW_HzSq)
            self.p_nl_c2 = c2**0.5

            if c1 >= self.max_BW_Hz:
                ret = False
                c1 = self.max_BW_Hz
            else:
                c1 = c1 / (1 - c1 / self.max_BW_Hz)
            self.p_nl_c1 = c1**0.5
            return ret
        return

    def roots(self):
        # second order sections (2x roots either real or complex conj)
        c1 = self.p_nl_c1 ** 2
        D1 = c1 / self.max_BW_Hz + 1
        c1 = c1 / D1
        if self.single_root:
            if not self.unstable:
                return [-c1 - self.min_BW_Hz]
            else:
                return [c1 + self.min_BW_Hz]
        else:
            c2 = self.p_nl_c2 ** 2
            D2 = c2 / self.max_BW_HzSq + 1
            c2 = c2 / D2
            # a = c2, b = c1, c = 1
            disc = c1 * c1 - 4 * c2
            if disc >= 0:
                sqrt_disc = disc ** 0.5
                if c1 < 0:
                    r1 = (-c1 + sqrt_disc) / 2
                else:
                    r1 = (-c1 - sqrt_disc) / 2
                if c2 == 0 and r1 == 0:
                    r2 = 0
                else:
                    r2 = c2 / r1
                if not self.unstable:
                    return [r1 - self.min_BW_Hz, r2 - self.min_BW_Hz]
                else:
                    return [-r1 + self.min_BW_Hz, -r2 + self.min_BW_Hz]
            else:
                sqrt_disc = (-disc) ** 0.5
                if not self.unstable:
                    r1 = (-c1 + sqrt_disc * 1j) / 2 - self.min_BW_Hz
                else:
                    r1 = (+c1 + sqrt_disc * 1j) / 2 + self.min_BW_Hz
                return [r1, r1.conjugate()]

    def roots_r(self):
        # second order sections (2x roots either real or complex conj)
        c1 = self.p_nl_c1 ** 2
        D1 = c1 / self.max_BW_Hz + 1
        c1 = c1 / D1
        if self.single_root:
            if not self.unstable:
                return [-c1 - self.min_BW_Hz]
            else:
                return [c1 + self.min_BW_Hz]
        else:
            c2 = self.p_nl_c2 ** 2
            D2 = c2 / self.max_BW_HzSq + 1
            c2 = c2 / D2
            # a = c2, b = c1, c = 1
            disc = c1 * c1 - 4 * c2
            if disc >= 0:
                sqrt_disc = disc ** 0.5
                if c1 < 0:
                    r1 = (-c1 + sqrt_disc) / 2
                else:
                    r1 = (-c1 - sqrt_disc) / 2

                if c2 == 0 and r1 == 0:
                    r2 = 0
                else:
                    r2 = c2 / r1

                if not self.unstable:
                    return [r1 - self.min_BW_Hz, r2 - self.min_BW_Hz]
                else:
                    return [-r1 + self.min_BW_Hz, -r2 + self.min_BW_Hz]
            else:
                return []

    def roots_c(self):
        # second order sections (2x roots either real or complex conj)
        c1 = self.p_nl_c1 ** 2
        D1 = c1 / self.max_BW_Hz + 1
        c1 = c1 / D1
        if self.single_root:
            return []
        else:
            c2 = self.p_nl_c2 ** 2
            D2 = c2 / self.max_BW_HzSq + 1
            c2 = c2 / D2
            # a = c2, b = c1, c = 1
            disc = c1 * c1 - 4 * c2
            if disc >= 0:
                return []
            else:
                sqrt_disc = (-disc) ** 0.5
                if not self.unstable:
                    r1 = (-c1 + sqrt_disc * 1j) / 2 - self.min_BW_Hz
                else:
                    r1 = (+c1 + sqrt_disc * 1j) / 2 + self.min_BW_Hz
                return [r1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


class PolyConstraints(object):
    """
    Constraint flags used for polynomials. These have correspondences
    for the root constraints for most polynomial classes
    """

    no_constraint = frozenset()
    even_real = frozenset(["even_real"])
    odd_real = frozenset(["odd_real"])
    odd_imag = frozenset(["odd_imag"])
    palendromic = frozenset(["palendromic"])
    odd_zero = odd_imag | odd_real
    eRoR = even_real | odd_real
    eRoI = even_real | odd_imag
    eRoZ = even_real | odd_zero
    eRoRpal = even_real | odd_real | palendromic
    eRoIpal = even_real | odd_imag | palendromic
    eRoZpal = even_real | odd_zero | palendromic


poly_constraints = PolyConstraints()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..codings_cmn import (
    CodingType,
)


class CodingRealnlBW(CodingType):
    N_parameters = 1
    unstable = False
    p_nl_BW = 0
    min_BW_Hz = 0

    def update(self, nl_BW):
        self.p_nl_BW = nl_BW

    def reduce(self):
        return [self.p_nl_BW]

    def update_roots(self, r1):
        assert r1.imag == 0
        self.check_dist_limit(thresh=1)

        if r1.real <= 0:
            self.unstable = False
            r1p = -r1.real
        else:
            self.unstable = True
            r1p = r1.real

        if r1p > 2 * self.min_BW_Hz:
            self.p_nl_BW = (r1p - self.min_BW_Hz) ** 0.5
            return True
        else:
            self.p_nl_BW = self.min_BW_Hz ** 0.5
            return False

    @property
    def gain_effect(self):
        if self.unstable:
            return -1
        else:
            return 1

    @property
    def F_Hz(self):
        return 0

    @property
    def BW_Hz(self):
        return self.min_BW_Hz + self.p_nl_BW ** 2

    def check_dist_limit(self, thresh=1):
        if self.sys.distance_limit_auto >= thresh:
            self.min_BW_Hz = self.sys.distance_limit(0)
            return True
        return False

    def transfer(self):
        self.check_dist_limit(thresh=2)
        # frequency, logarithmic amplitude
        if not self.unstable:
            r = -self.BW_Hz
        else:
            r = self.BW_Hz
        X = self.sys.Xsf_grid
        return X - r

    def derivative(self):
        self.check_dist_limit(thresh=2)
        # real/imaginary part of root
        if not self.unstable:
            DrDp = 2 * self.p_nl_BW
        else:
            DrDp = -2 * self.p_nl_BW

        return [DrDp]

    def roots_r(self):
        if not self.unstable:
            r = -self.BW_Hz
        else:
            r = self.BW_Hz
        return [r]



# -*- coding: utf-8 -*-
"""
"""
import numpy as np
from IIRrational_test_data import IIRrational_data_dev
from wield.iirrational.utilities.mpl import mplfigB
from wield.iirrational import testing

from wield.iirrational.fitters_ZPK import ZPKrep2MRF
from wield.iirrational.fitters_ZPK import codings_s
from wield.iirrational.fitters_ZPK.MRF import check_jac

from utilities import check_residuals


def prep_fitter():
    data = IIRrational_data_dev("quad_M0_P2V_10Hz")

    from wield.iirrational import fitters_rational

    SNR = np.minimum(data.SNR, 100)
    fitter = fitters_rational.ChebychevFilter(
        F_Hz=data.F_Hz,
        data=data.data,
        W=SNR,
        npoles=100,
        nzeros=100,
    )
    # fitter.poles = (.90,) * 0
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_SVD()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    # fitter.matched_pairs_clear()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.stabilize = True
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.matched_pairs_clear()
    # axB = plot_fit(fitter, with_error=False, plot_zp = True)
    # axB = mplfigB()
    # axB.ax0.semilogy(fitter.F_Hz, abs(fitter.xfer_fit))
    return fitter


def test_NL_fitter1():
    fitter = prep_fitter()
    rep_s = ZPKrep2MRF(
        fitter.ZPKrep,
        coding_map=codings_s.coding_maps.nlFBW_safe,
    )

    print(rep_s.residuals_average)
    # axB = mplfigB()

    check_residuals(rep_s)
    opt = rep_s.optimize()
    print(rep_s.residuals_average)
    assert rep_s.residuals_average < 200

    # axB = plot_fitter_flag(rep_s, with_error=False, plot_zp = True)



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

# from wield.iirrational import fitters
from wield.iirrational import fitters_rational
from wield.iirrational import plots
from wield.iirrational.utilities import relpy
import numpy as np

try:
    from IIRrational_test_data import IIRrational_data_dev
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_disc_convergence():
    data = IIRrational_data_dev("quad_M0_P2V_10Hz")
    SNR = np.minimum(data.SNR, 100)
    F_nyquist_Hz = np.max(data.F_Hz)
    fitter = fitters_rational.RationalDiscFilter(
        F_Hz=data.F_Hz,
        data=data.data,
        W=SNR,
        F_nyquist_Hz=F_nyquist_Hz,
        npoles=100,
        nzeros=100,
    )
    print("data", fitter.data)
    print("W", fitter.W)
    print("h_a", fitter.h_a.xfer, fitter.h_a.lnG)
    print("h_b", fitter.h_b.xfer, fitter.h_b.lnG)
    fitter.fit_zeros()
    fitter.fit_poles()
    print("h_a", fitter.h_a.xfer[0], fitter.h_a.lnG)
    print("h_b", fitter.h_b.xfer[0], fitter.h_b.lnG)
    fitter.fit_zeros()
    print("h_a", fitter.h_a.xfer[0], fitter.h_a.lnG)
    print("h_b", fitter.h_b.xfer[0], fitter.h_b.lnG)
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.3)
    fitter.stabilize = True
    # fitter.npoles = len(fitter.poles)
    # fitter.nzeros = len(fitter.zeros)
    ##fitter.stabilize = True
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.7)

    # TODO
    #!Make this a TEST so it doesn't plot all of the time

    axB = plots.plot_fit(
        fitter,
        plot_zp=True,
        # plot_past_data = False,
        plot_past_data=True,
        xscale="log",
    )

    axB.save(relpy(__file__, "plots/disc-convergence"))
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np

from .algorithms import (
    cheby_sequence,
    disc_sequence,
)



def fit_cheby(aid, order_hint=None):
    if order_hint is None:
        return fit_cheby_base(
            aid,
            order=aid.hint(
                "rational_cheby_fit_order",
                "rational_fit_order",
                "order_initial",
            ),
            order_max=aid.hint(
                "rational_cheby_fit_order_max",
                "rational_fit_order_max",
                "order_max",
            ),
            order_min=aid.hint(
                "rational_cheby_fit_order_min",
                "rational_fit_order_min",
                "order_min",
            ),
        )
    else:
        return fit_cheby_base(
            aid,
            order=aid.hint(
                order_hint,
                "rational_cheby_fit_order",
                "rational_fit_order",
                "order_initial",
            ),
            order_max=aid.hint(
                "rational_cheby_fit_order_max",
                "rational_fit_order_max",
                "order_max",
            ),
            order_min=aid.hint(
                "rational_cheby_fit_order_min",
                "rational_fit_order_min",
                "order_min",
            ),
        )


def fit_cheby_base(
    aid,
    order,
    order_max,
    order_min,
):
    if order == 0:
        return

    aid.log_progress(4, "chebychev rational fit")
    # rat_fitter = v2.disc_sequence.rational_disc_fit(
    reldeg = aid.hint("relative_degree")
    if reldeg is None:
        reldeg_max = aid.hint("relative_degree_max")
        reldeg_min = aid.hint("relative_degree_min")
        if reldeg_min is None and reldeg_max is None:
            reldeg = None
        elif reldeg_min is None:
            reldeg = 0
        elif reldeg_max is None:
            reldeg = 0
        else:
            reldeg = int((reldeg_min + reldeg_max) // 2)

    factor_orders = aid.fitter_orders("factors")
    if reldeg is not None:
        diff_reldeg = reldeg - factor_orders.reldeg
    else:
        diff_reldeg = 0

    factors_order = aid.fitter_orders().factors_maxzp
    print("RELDEG: ", factor_orders.reldeg, aid.fitter_orders().factors_reldeg, diff_reldeg)
    # print("FACTORS ORDER", factors_order)
    if order is not None:
        order -= factors_order
        order = max(order, 4)

    rat_fitter = cheby_sequence.rational_cheby_fit(
        aid=aid,
        order_max=order_max - factors_order,
        order_min=max(order_min - factors_order, 2),
        order=order,
        relative_degree=diff_reldeg,
        ZPKrep=aid.fitter.ZPKrep,
    )

    fitter = aid.fitter.regenerate(
        ZPKrep=rat_fitter.ZPKrep,
    )

    assert(np.all(aid.fitter.poles.fullplane.real < 0))

    aid.fitter_update(
        fitter,
        representative=False,
    )


def fit_disc_factored(aid):
    with aid.factorization():
        fit_disc(aid)


def fit_disc(aid):
    aid.log_progress(4, "rational disc fitting")
    order = aid.hint(
        "rational_cheby_fit_order",
        "rational_fit_order",
        "order_initial",
    )
    order_max = aid.hint(
        "rational_cheby_fit_order_max",
        "rational_fit_order_max",
        "order_max",
    )
    order_min = aid.hint(
        "rational_cheby_fit_order_min",
        "rational_fit_order_min",
        "order_min",
    )

    # TODO
    from wield.iirrational.v1 import disc_sequence

    rat_fitter = disc_sequence.rational_disc_fit(
        F_Hz=aid.fitter.F_Hz,
        data=aid.fitter.data,
        nyquist_final_Hz=None,
        SNR=aid.fitter.W,
        order=order,
        order_max=order_max,
    )
    # rat_fitter = disc_sequence.rational_disc_fit(
    #    aid             = aid,
    #    order_max       = order_max,
    #    order_min       = order_min,
    #    order           = order,
    #    ZPKrep          = aid.fitter.ZPKrep,
    # )

    fitter = aid.fitter.regenerate(
        ZPKrep=rat_fitter.ZPKrep,
    )

    # TODO
    # validate = aid.hint('rational_fitter_validate')
    # if validate is not None:
    #    validate(rat_fitter, fitter)

    aid.fitter_update(
        fitter,
        representative=False,
    )

    aid.log(
        1,
        "Initial Order: (Z={0}, P={1}, Z-P={2})".format(
            len(fitter.zeros),
            len(fitter.poles),
            len(fitter.zeros) - len(fitter.poles),
        ),
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import warnings
import os
import contextlib
from os import path

from wield import declarative
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
import numpy as np
from wield.bunch import Bunch, DeepBunch

from .colors import color_array

try:
    org_mode
except NameError:
    org_mode = False


def save_figure_MP(fig, fname, *args, **kwargs):
    """
    After pickling to a subprocess, the canvas is destroyed due to matplotlib bullshit, so make a new one as it apparently doesn't
    do this for you
    """
    if fig.canvas is None:
        canvas = FigureCanvas(fig)
        fig.set_canvas(canvas)
    return fig.savefig(fname, *args, **kwargs)


class SaveToken(declarative.OverridableObject):
    aps = None
    fbasename = None
    kwargs = {}

    def __lshift__(self, other):
        self.aps(self.fbasename, fig_or_fbunch=other, **self.kwargs)
        return other

    def __rlshift__(self, other):
        self.aps(self.fbasename, fig_or_fbunch=other, **self.kwargs)
        return other

    def __rshift__(self, other):
        self.aps(self.fbasename, fig_or_fbunch=other, **self.kwargs)
        return other

    def __rrshift__(self, other):
        self.aps(self.fbasename, fig_or_fbunch=other, **self.kwargs)
        return other


def mpl_autorasterize(fig):
    children_current = fig.get_children()
    children_ever = set()
    while children_current:
        child = children_current.pop()
        if child in children_ever:
            continue
        else:
            children_ever.add(child)

        try:
            more_children = child.get_children()
        except AttributeError:
            pass
        else:
            children_current.extend(more_children)

        try:
            xdat = child.get_xdata()
            if len(xdat) > 100:
                child.set_rasterized(True)
                child.set_antialiased(True)
            continue
        except AttributeError:
            pass

        try:
            paths = child.get_paths()
            for p in paths:
                if len(p.vertices) > 100:
                    child.set_rasterized(True)
                    child.set_antialiased(True)
                    # print(child, len(xdat))
        except AttributeError:
            pass


class AutoPlotSaver(declarative.OverridableObject):
    max_width_in = None
    max_height_in = None
    save_dpi = 400

    org_dpi = 100
    org_subfolder = None

    rasterize_auto = True

    formats = DeepBunch()
    formats.pdf.use = True
    formats.jpg.use = False
    formats.jpg.dpi = 200
    formats.jpg.facecolorize = True
    formats.png.facecolorize = True
    formats.png.use = False

    embed = False
    save_show = True
    fixname = False
    _pool = None

    _last_async_result = None

    @contextlib.contextmanager
    def pool(self, workers=4):
        """
        runs the plot save in a contextmanager and waits for the plotting to be done simultaneously
        """
        import multiprocessing

        wasnone = False
        if self._pool is None:
            if workers > 1:
                asavefig._pool = multiprocessing.Pool(workers)
                asavefig._last_async_result = []
            wasnone = True
        yield
        if asavefig._last_async_result is not None:
            for result in asavefig._last_async_result:
                result.get()
            asavefig._last_async_result = []
        if wasnone:
            if asavefig._pool is not None:
                asavefig._pool.close()
                asavefig._pool.join()
                asavefig._pool = None
            asavefig._last_async_result = None

    def __call__(
        self,
        fbasename,
        fig_or_fbunch=None,
        fixname=None,
    ):

        if fig_or_fbunch is None:
            return SaveToken(
                aps=self,
                fbasename=fbasename,
                kwargs=dict(
                    fixname=fixname,
                ),
            )

        fixname = fixname if fixname is not None else self.fixname

        try:
            fig = fig_or_fbunch.fig

            formats = fig_or_fbunch.get("formats", None)
            # the "get" method unwraps the deepbunch
            formats = DeepBunch(formats)
            if not formats:
                formats = self.formats

            save_show = fig_or_fbunch.get("save_show", None)
            # and needed since show may be empty DeepBunch
            if not save_show and save_show is not False:
                save_show = self.save_show

        except AttributeError:
            fig = fig_or_fbunch
            save_show = self.save_show
            formats = self.formats
        w, h = fig.get_size_inches()
        if self.max_width_in is not None and w > self.max_width_in:
            new_w = self.max_width_in
            new_h = float(h) / float(w) * new_w
            fig.set_size_inches(new_w, new_h)

        # this silly bit reduces the formats to only the one specified
        fbase, fext = path.splitext(fbasename)
        if fext:
            fbasename = fbase
            # cut off the dot
            fext = fext[1:]
            single_formats = DeepBunch()
            # apply any settings stored in this object or the plot itself
            single_formats[fext].update_recursive(self.formats[fext])
            single_formats[fext].update_recursive(formats[fext])
            # force usage of this single format!
            single_formats[fext].use = True
            formats = single_formats

        if self.rasterize_auto:
            mpl_autorasterize(fig)

        subfolder = ""
        if self.org_subfolder:
            subfolder = self.org_subfolder

        fbasepath, fbasefname = path.split(fbasename)
        if "_" in fbasefname and fixname:
            warnings.warn(
                "Image name contains '_' which will be changed to '-' to fix nbsphinx export"
            )
            fbasefname = fbasefname.replace("_", "-")
            fbasename = path.join(fbasepath, fbasename)

        fbasename = path.join(subfolder, fbasename)
        dirname = path.dirname(fbasename)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

        global org_mode
        used_png = False
        for fmt, fB in formats.items():
            if fmt == "png" and (org_mode or self.org_subfolder):
                # to avoide the elif
                pass
            elif not fB.use:
                continue
            if fmt == "png":
                used_png = True
            if fB.dpi:
                dpi = fB.dpi
            else:
                dpi = self.save_dpi
            kwargs = dict()
            if fB.facecolorize:
                kwargs["facecolor"] = fig.get_facecolor()
            if self._pool is None:
                fig.savefig(
                    fbasename + "." + fmt,
                    dpi=dpi,
                    bbox_inches="tight",
                    # tight_layout=True,
                    pad_inches=0.05,
                    transparent=True,
                    quality=50,
                    **kwargs
                )
            else:
                mydir = os.getcwd()
                self._last_async_result.append(
                    self._pool.apply_async(
                        save_figure_MP,
                        args=(
                            fig,
                            os.path.join(mydir, fbasename + "." + fmt),
                        ),
                        kwds=dict(
                            dpi=dpi,
                            bbox_inches="tight",
                            pad_inches=0.05,
                            transparent=True,
                            quality=50,
                            # tight_layout =True,
                            **kwargs
                        ),
                    )
                )

        if used_png:
            fname = fbasename + ".png"
            if org_mode:
                print("figure: {0}".format(fname))
                print("[[file:{0}]]".format(fname))
            if not self.embed:
                if save_show:
                    try:
                        import IPython.display
                        import time

                        # IPython.display.display(IPython.display.Image(filename=fname, embed=False))
                        # html_bit = '<img src="{1}/../{0}?{1}">'.format(fname, int(time.time()))
                        # IPython.display.display(IPython.display.HTML(html_bit))
                        ftype_md = []
                        for fmt, fB in formats.items():
                            if fB.use or fmt == "png":
                                md = "[{ftype}]({fbasename}.{ftype})".format(
                                    ftype=fmt, fbasename=fbasename
                                )
                                ftype_md.append(md)
                        markdown_bit = '![{fbasename}]({0}?{1} "{fbasename}")'.format(
                            fname,
                            int(time.time()),
                            fbasename=fbasename,
                        )
                        IPython.display.display(
                            IPython.display.Markdown(
                                markdown_bit + "\n" + ",  ".join(ftype_md)
                            )
                        )
                        plt.close(fig)
                    except ImportError:
                        pass
                else:
                    plt.close(fig)
            else:
                if save_show:
                    try:
                        import IPython.display
                        import time

                        IPython.display.display(
                            IPython.display.Image("{0}".format(fname))
                        )
                        plt.close(fig)
                    except ImportError:
                        pass
                else:
                    plt.close(fig)
        fig.set_dpi(mpl.rcParams["figure.dpi"])
        return


asavefig = AutoPlotSaver()


def patchify_axes(ax, plotname, check_log_Y=False):
    oldplot = getattr(ax, plotname)

    def plot(X, Y, *args, **kwargs):
        Y = np.asarray(Y)
        b = np.broadcast(X, Y)

        if check_log_Y and np.all(Y <= 0):
            return

        if b.shape != Y.shape:
            Y = np.ones(X.shape) * Y
        return oldplot(X, Y, *args, **kwargs)

    plot.__name__ = oldplot.__name__
    plot.__doc__ = oldplot.__doc__
    setattr(ax, plotname, plot)


def patch_axes(ax):
    patchify_axes(ax, "plot")
    patchify_axes(ax, "loglog", check_log_Y=True)
    patchify_axes(ax, "semilogy", check_log_Y=True)
    patchify_axes(ax, "semilogx")


def mplfigB(
    Nrows=1,
    Ncols=1,
    size_in=(None, None),
    size_in_base=(None, None),
    size_in_dW_dH=(3, 1),
    x_by_col=False,
    prop_cycle=None,
):
    if isinstance(Nrows, (list, tuple)):
        rownames = Nrows
        Nrows = len(rownames)
    else:
        rownames = None

    width_in, height_in = size_in
    size_in_base_W, size_in_base_H = size_in_base

    if size_in_base_W is None:
        size_in_base_W = mpl.rcParams["figure.figsize"][0]

    if size_in_base_H is None:
        size_in_base_H = mpl.rcParams["figure.figsize"][1]

    if width_in is None:
        width_in = size_in_base_W + Ncols * size_in_dW_dH[0]

    if height_in is None:
        height_in = size_in_base_H + Nrows * size_in_dW_dH[1]

    axB = Bunch()
    axB.fig = plt.figure()
    axB.fig.set_size_inches(width_in, height_in)
    attach_finalizer(axB)

    global asavefig

    def save(rootname, **kwargs):
        axB.finalize()
        axB << asavefig(rootname, **kwargs)

    axB.save = save

    N = 0
    axB.ax_grid_colrow = []
    for idx_col in range(Ncols):
        ax_list = []
        axB.ax_grid_colrow.append([])
        for idx_row in range(Nrows):
            if x_by_col:
                if idx_row != 0:
                    sharex = axB.ax_grid_colrow[idx_col][0]
                else:
                    sharex = None
            else:
                sharex = None
            ax = axB.fig.add_subplot(
                Nrows, Ncols, idx_row * Ncols + idx_col + 1, sharex=sharex
            )
            if prop_cycle is not None:
                ax.set_prop_cycle(color=prop_cycle)
            # patch_axes(ax)
            ax_list.append(ax)
            ax.grid(b=True)
            ax.grid(b=True, which="minor", color=(0.9, 0.9, 0.9), lw=0.5)
            axB.ax_grid_colrow[idx_col].append(ax)
            axB["ax{0}_{1}".format(idx_row, idx_col)] = ax
            axB["ax{0}".format(N)] = ax

            if rownames is not None:
                axB[rownames[N]] = ax

            N += 1
            if idx_col == 0:
                if idx_row == 0:
                    axB.ax_top = ax
                if idx_row == Nrows - 1:
                    axB.ax_bottom = ax
                axB["ax_list"] = ax_list
        axB["ax_list_{0}".format(idx_col)] = ax_list
    return axB


def attach_finalizer(ax):
    ax.finalizers = []
    ax.finalized = False

    def finalize():
        ax.finalized = True
        for f in ax.finalizers:
            f()

    ax.finalize = finalize
    return


asavefig = AutoPlotSaver()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from ..codings_cmn import (
    CodingTypeZ,
)


# TODO, fix this class
class CodingGainNLDelay(CodingTypeZ):
    N_parameters = 2
    max_delay_samples = 100
    min_delay_samples = -1
    p_nl_delay = -1000

    # delay_samples = min_delay_samples + (1 + p_nl_delay / (1. + p_nl_delay**2)**.5) * ((max_delay_samples - min_delay_samples)/2)
    # (delay_samples - min_delay_samples) / ((max_delay_samples - min_delay_samples)/2) = X = (1 + p_nl_delay / (1. + p_nl_delay**2)**.5)
    # (X-1)**2 = p_nl_delay**2 / (1 + p_nl_delay**2)
    # (X-1)**2 = (1 - (X - 1)**2) * p_nl_delay**2
    # (X-1)**2 / (1 - (X - 1)**2) = p_nl_delay**2
    __X = (0 - min_delay_samples) / ((max_delay_samples - min_delay_samples) / 2)
    p_nl_delay = -(((__X - 1) ** 2 / (1 - (__X - 1) ** 2)) ** 0.5)

    del __X  # delete the temporary so it isn't in the class

    def setup(
        self,
        delay_s=None,
    ):
        if delay_s is not None:
            # self.p_delay = delay_s
            # doesn't actually know the nyqyist, so cant get samples, yet
            # __X = (0 - self.min_delay_samples) / ((self.max_delay_samples - self.min_delay_samples)/2)
            # p_nl_delay = -((__X-1)**2 / (1 - (__X - 1)**2))**.5
            pass

        return

    def update(self, nl_delay):
        self.p_nl_delay = nl_delay

    def reduce(self):
        return [self.p_nl_delay]

    @property
    def delay_samples(self):
        delay_samples = self.min_delay_samples + (
            1 + self.p_nl_delay / (1.0 + self.p_nl_delay ** 2) ** 0.5
        ) * ((self.max_delay_samples - self.min_delay_samples) / 2)
        return delay_samples

    @property
    def delay_s(self):
        self.delay_samples * self.F_nyquist_Hz

    def delay_set(self, delay_s):
        self.F_nyquist_Hz = self.sys.F_nyquist_Hz
        assert False

    def transfer(self):
        self.F_nyquist_Hz = self.sys.F_nyquist_Hz
        # gain term (can only be 1 and at the start)
        return self.p_gain * np.exp(
            -Ipi * self.delay_samples / self.sys.F_nyquist_Hz * self.sys.F_Hz
        )

    def derivative(self):
        self.F_nyquist_Hz = self.sys.F_nyquist_Hz
        sqp5 = (1.0 + self.p_nl_delay ** 2) ** 0.5
        dDelaydNL = ((self.max_delay_samples - self.min_delay_samples) / 2) * (
            1 / sqp5 - self.p_nl_delay ** 2 / (1.0 + self.p_nl_delay ** 2) ** 1.5
        )
        return [
            dDelaydNL
            * -Ipi
            * self.sys.F_Hz
            / self.sys.F_nyquist_Hz
            * np.exp(-Ipi * self.delay_samples * self.sys.F_Hz / self.sys.F_nyquist_Hz)
            if not self.lock_delay
            else 0,
        ]

    def roots(self):
        return []
